from typing import TYPE_CHECKING, Optional

from polyaxon import settings
from polyaxon.constants.globals import NO_AUTH
from polyaxon.sdk.api import (
    AgentsV1Api,
    AuthV1Api,
    ConnectionsV1Api,
    DashboardsV1Api,
    OrganizationsV1Api,
    PresetsV1Api,
    ProjectDashboardsV1Api,
    ProjectSearchesV1Api,
    ProjectsV1Api,
    QueuesV1Api,
    RunsV1Api,
    SearchesV1Api,
    ServiceAccountsV1Api,
    TagsV1Api,
    TeamsV1Api,
    UsersV1Api,
    VersionsV1Api,
)
from polyaxon.sdk.async_client.api_client import AsyncApiClient
from polyaxon.sdk.sync_client.api_client import ApiClient

if TYPE_CHECKING:
    from polyaxon.schemas.cli.client_config import ClientConfig


class PolyaxonClient:
    """Auto-configurable and high-level base client that abstract
    the need to set a configuration for each service.

    PolyaxonClient comes with logic
    to pass config and token to other specific clients.

    If no values are passed to this class,
    Polyaxon will try to resolve the configuration from the environment:
     * If you have a configured CLI, Polyaxon will use the configuration of the cli.
     * If you use this client in the context of a job or a service managed by Polyaxon,
       a configuration will be available.

    Args:
        config: ClientConfig, optional, Instance of a ClientConfig.
        token: str, optional, the token to use for authenticating the clients,
               if the user is already logged in using the CLI, it will automatically use that token.
               Using the client inside a job/service scheduled with Polyaxon will have access to the
               token of the user who started the run if the `auth` context is enabled.

    You can access specific low level clients:

    ```python
    >>> client = PolyaxonClient()

    >>> client.projects_v1
    >>> client.runs_v1
    >>> client.auth_v1
    >>> client.users_v1
    >>> client.agents_v1
    >>> client.connections_v1
    >>> client.organizations_v1
    ```

    If you are interacting with a run or with a project, we suggest that you check:
     * [RunClient](/docs/core/python-library/run-client/)
     * [ProjectClient](/docs/core/python-library/project-client/)
    """

    def __init__(
        self,
        config: Optional["ClientConfig"] = None,
        token: Optional[str] = None,
        is_async: bool = False,
    ):
        self._config = config or settings.CLIENT_CONFIG
        token = token or self._config.token
        if not token and settings.AUTH_CONFIG:
            self._config.token = settings.AUTH_CONFIG.token
        elif token == NO_AUTH:
            self._config.token = None
        else:
            self._config.token = token

        self.is_async = is_async
        self.api_client = self._get_client()
        self._projects_v1 = None
        self._runs_v1 = None
        self._project_dashboards_v1 = None
        self._project_searches_v1 = None
        self._auth_v1 = None
        self._users_v1 = None
        self._versions_v1 = None
        self._agents_v1 = None
        self._queues_v1 = None
        self._service_accounts_v1 = None
        self._presets_v1 = None
        self._tags_v1 = None
        self._teams_v1 = None
        self._connections_v1 = None
        self._dashboards_v1 = None
        self._searches_v1 = None
        self._organizations_v1 = None

    def _get_client(self):
        if self.is_async:
            return AsyncApiClient(
                self.config.async_sdk_config, **self.config.client_header
            )
        return ApiClient(self.config.sdk_config, **self.config.client_header)

    def reset(self):
        self._projects_v1 = None
        self._runs_v1 = None
        self._project_dashboards_v1 = None
        self._project_searches_v1 = None
        self._auth_v1 = None
        self._users_v1 = None
        self._versions_v1 = None
        self._agents_v1 = None
        self._queues_v1 = None
        self._service_accounts_v1 = None
        self._presets_v1 = None
        self._tags_v1 = None
        self._teams_v1 = None
        self._connections_v1 = None
        self._dashboards_v1 = None
        self._searches_v1 = None
        self._organizations_v1 = None
        self.api_client = self._get_client()

    @property
    def config(self):
        return self._config

    @property
    def projects_v1(self):
        if not self._projects_v1:
            self._projects_v1 = ProjectsV1Api(self.api_client)
        return self._projects_v1

    @property
    def runs_v1(self):
        if not self._runs_v1:
            self._runs_v1 = RunsV1Api(self.api_client)
        return self._runs_v1

    @property
    def auth_v1(self):
        if not self._auth_v1:
            self._auth_v1 = AuthV1Api(self.api_client)
        return self._auth_v1

    @property
    def users_v1(self):
        if not self._users_v1:
            self._users_v1 = UsersV1Api(self.api_client)
        return self._users_v1

    @property
    def versions_v1(self):
        if not self._versions_v1:
            self._versions_v1 = VersionsV1Api(self.api_client)
        return self._versions_v1

    @property
    def agents_v1(self):
        if not self._agents_v1:
            self._agents_v1 = AgentsV1Api(self.api_client)
        return self._agents_v1

    @property
    def queues_v1(self):
        if not self._queues_v1:
            self._queues_v1 = QueuesV1Api(self.api_client)
        return self._queues_v1

    @property
    def service_accounts_v1(self):
        if not self._service_accounts_v1:
            self._service_accounts_v1 = ServiceAccountsV1Api(self.api_client)
        return self._service_accounts_v1

    @property
    def tags_v1(self):
        if not self._tags_v1:
            self._tags_v1 = TagsV1Api(self.api_client)
        return self._tags_v1

    @property
    def teams_v1(self):
        if not self._teams_v1:
            self._teams_v1 = TeamsV1Api(self.api_client)
        return self._teams_v1

    @property
    def connections_v1(self):
        if not self._connections_v1:
            self._connections_v1 = ConnectionsV1Api(self.api_client)
        return self._connections_v1

    @property
    def project_dashboards_v1(self):
        if not self._project_dashboards_v1:
            self._project_dashboards_v1 = ProjectDashboardsV1Api(self.api_client)
        return self._project_dashboards_v1

    @property
    def project_searches_v1(self):
        if not self._project_searches_v1:
            self._project_searches_v1 = ProjectSearchesV1Api(self.api_client)
        return self._project_searches_v1

    @property
    def dashboards_v1(self):
        if not self._dashboards_v1:
            self._dashboards_v1 = DashboardsV1Api(self.api_client)
        return self._dashboards_v1

    @property
    def searches_v1(self):
        if not self._searches_v1:
            self._searches_v1 = SearchesV1Api(self.api_client)
        return self._searches_v1

    @property
    def presets_v1(self):
        if not self._presets_v1:
            self._presets_v1 = PresetsV1Api(self.api_client)
        return self._presets_v1

    @property
    def organizations_v1(self):
        if not self._organizations_v1:
            self._organizations_v1 = OrganizationsV1Api(self.api_client)
        return self._organizations_v1

    def sanitize_for_serialization(self, value):
        return self.api_client.sanitize_for_serialization(value)



from typing import Dict, List, Optional, Union
from typing_extensions import Literal

from clipped.types.ref_or_obj import RefField
from clipped.utils.enums import PEnum
from pydantic import Field, StrictStr, validator

from polyaxon.k8s import k8s_schemas, k8s_validation
from polyaxon.polyflow.run.base import BaseRun
from polyaxon.polyflow.run.kinds import V1RunKind
from polyaxon.polyflow.run.resources import V1RunResources
from polyaxon.polyflow.run.spark.replica import V1SparkReplica
from polyaxon.polyflow.run.utils import DestinationImageMixin


class V1SparkType(str, PEnum):
    JAVA = "java"
    SCALA = "scala"
    PYTHON = "python"
    R = "r"


class V1SparkDeploy(str, PEnum):
    CLUSTER = "cluster"
    CLIENT = "client"
    IN_CLUSTER_CLIENT = "in_cluster_client"


class V1SparkJob(BaseRun, DestinationImageMixin):
    """Spark jobs are used to run Spark applications on Kubernetes.

    [Apache Spark](https://spark.apache.org/) is data-processing engine.

    Args:
        kind: str, should be equal `sparkjob`
        connections: List[str], optional
        volumes: List[[Kubernetes Volume](https://kubernetes.io/docs/concepts/storage/volumes/)],
             optional
        type: str [`JAVA`, `SCALA`, `PYTHON`, `R`]
        spark_version: str, optional
        python_version: str, optional
        deploy_mode: str, optional
        main_class: str, optional
        main_application_file: str, optional
        arguments: List[str], optional
        hadoop_conf: Dict[str, str], optional
        spark_conf: Dict[str, str], optional
        hadoop_config_map: str, optional
        spark_config_map: str, optional
        executor: [V1SparkReplica](/docs/experimentation/distributed/spark-replica/)
        driver: [V1SparkReplica](/docs/experimentation/distributed/spark-replica/)

    ## YAML usage

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   connections:
    >>>   volumes:
    >>>   type:
    >>>   sparkVersion:
    >>>   deployMode:
    >>>   mainClass:
    >>>   mainApplicationFile:
    >>>   arguments:
    >>>   hadoopConf:
    >>>   sparkConf:
    >>>   hadoopConfigMap:
    >>>   sparkConfigMap:
    >>>   executor:
    >>>   driver:
    ```

    ## Python usage

    ```python
    >>> from polyaxon.polyflow import V1Environment, V1Init, V1SparkJob, V1SparkReplica, V1SparkType
    >>> from polyaxon.k8s import k8s_schemas
    >>> spark_job = V1SparkJob(
    >>>     connections=["connection-name1"],
    >>>     volumes=[k8s_schemas.V1Volume(...)],
    >>>     type=V1SparkType.PYTHON,
    >>>     spark_version="3.0.0",
    >>>     spark_conf={...},
    >>>     driver=V1SparkReplica(...),
    >>>     executor=V1SparkReplica(...),
    >>> )
    ```

    ## Fields

    ### kind

    The kind signals to the CLI, client, and other tools that this component's runtime is a job.

    If you are using the python client to create the runtime,
    this field is not required and is set by default.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    ```

    ### connections

    A list of [connection names](/docs/setup/connections/) to resolve for the job.

    <blockquote class="light">
    If you are referencing a connection it must be configured.
    All referenced connections will be checked:

     * If they are accessible in the context of the project of this run

     * If the user running the operation can have access to those connections
    </blockquote>

    After checks, the connections will be resolved and inject any volumes, secrets, configMaps,
    environment variables for your main container to function correctly.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   connections: [connection1, connection2]
    ```

    ### volumes

    A list of [Kubernetes Volumes](https://kubernetes.io/docs/concepts/storage/volumes/)
    to resolve and mount for your jobs.

    This is an advanced use-case where configuring a connection is not an option.

    When you add a volume you need to mount it manually to your container(s).

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   volumes:
    >>>     - name: volume1
    >>>       persistentVolumeClaim:
    >>>         claimName: pvc1
    >>>   ...
    ```

    ### type

    Tells the type of the Spark application, possible values: `Java`, `Scala`, `Python`, `R`

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   type: Python
    >>>   ...
    ```

    ### sparkVersion

    The version of Spark the application uses.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   sparkVersion: 3.0.0
    >>>   ...
    ```

    ### deployMode

    The deployment mode of the Spark application.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   deployMode: cluster
    >>>   ...
    ```

    ### mainClass

    The fully-qualified main class of the Spark application.
    This only applies to Java/Scala Spark applications.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   mainClass: ...
    >>>   ...
    ```

    ### mainApplicationFile

    The path to a bundled JAR, Python, or R file of the application.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   mainApplicationFile: ...
    >>>   ...
    ```

    ### arguments

    List of arguments to be passed to the application.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   arguments: [...]
    >>>   ...
    ```

    ### hadoopConf

    HadoopConf carries user-specified Hadoop configuration properties as they would use
    the "--conf" option in spark-submit.
    The SparkApplication controller automatically adds prefix "spark.hadoop."
    to Hadoop configuration properties.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   hadoopConf: {...}
    >>>   ...
    ```

    ### sparkConf

    Carries user-specified Spark configuration properties as they would use the  "--conf" option in
    spark-submit.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   sparkConf: {...}
    >>>   ...
    ```

    ### hadoopConfigMap

    Carries the name of the ConfigMap containing Spark configuration files such as log4j.properties.
    The controller will add environment variable SPARK_CONF_DIR to
    the path where the ConfigMap is mounted to.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   hadoopConfigMap: {...}
    >>>   ...
    ```

    ### sparkConfigMap

    Carries the name of the ConfigMap containing Spark configuration files such as log4j.properties.
    The controller will add environment variable SPARK_CONF_DIR to the
    path where the ConfigMap is mounted to.

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   sparkConfigMap: {...}
    >>>   ...
    ```

    ### executor

    Executor is a spark replica specification

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   executor:
    >>>     replicas: 1
    >>>     ...
    >>>   ...
    ```

    ### driver

    Driver a spark replica specification

    ```yaml
    >>> run:
    >>>   kind: sparkjob
    >>>   driver:
    >>>     replicas: 1
    >>>     ...
    >>>   ...
    ```
    """

    _IDENTIFIER = V1RunKind.SPARKJOB
    _SWAGGER_FIELDS = [
        "volumes",
    ]

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    connections: Optional[Union[List[StrictStr], RefField]]
    volumes: Optional[Union[List[k8s_schemas.V1Volume], RefField]]
    type: Optional[V1SparkType]
    spark_version: Optional[StrictStr] = Field(alias="sparkVersion")
    python_version: Optional[StrictStr] = Field(alias="pythonVersion")
    deploy_mode: Optional[V1SparkDeploy] = Field(alias="deployMode")
    main_class: Optional[StrictStr] = Field(alias="mainClass")
    main_application_file: Optional[StrictStr] = Field(alias="mainApplicationFile")
    arguments: Optional[Union[List[StrictStr], RefField]]
    hadoop_conf: Optional[Union[Dict[StrictStr, StrictStr], RefField]] = Field(
        alias="hadoopConf"
    )
    spark_conf: Optional[Union[Dict[StrictStr, StrictStr], RefField]] = Field(
        alias="sparkConf"
    )
    hadoop_config_map: Optional[Union[Dict[StrictStr, StrictStr], RefField]] = Field(
        alias="hadoopConfigMap"
    )
    spark_config_map: Optional[Union[Dict[StrictStr, StrictStr], RefField]] = Field(
        alias="sparkConfigMap"
    )
    executor: Optional[Union[V1SparkReplica, RefField]]
    driver: Optional[Union[V1SparkReplica, RefField]]

    @validator("volumes", always=True, pre=True)
    def validate_volumes(cls, v):
        if not v:
            return v
        return [k8s_validation.validate_k8s_volume(vi) for vi in v]

    def get_resources(self):
        resources = V1RunResources()
        for i in range(self.replicas or 1):
            resources += V1RunResources.from_container(self.container)

        return resources

    def get_all_containers(self):
        return [self.container] if self.container else []

    def get_all_connections(self):
        return self.connections or []

    def get_all_init(self):
        return self.init or []



import os

from typing import Type

from polyaxon.config.manager import ConfigManager
from polyaxon.config.spec import ConfigSpec
from polyaxon.schemas.cli.client_config import ClientConfig


class ClientConfigManager(ConfigManager):
    """Manages client configuration .client file."""

    VISIBILITY = ConfigManager.Visibility.GLOBAL
    CONFIG_FILE_NAME = ".client"
    CONFIG: Type[ClientConfig] = ClientConfig

    @classmethod
    def get_config_from_env(cls, **kwargs) -> ClientConfig:
        tmp_path = cls.get_tmp_config_path()
        glob_path = cls.get_global_config_path()

        config = cls._CONFIG_READER.read_configs(
            [
                ConfigSpec(tmp_path, config_type=".json", check_if_exists=False),
                ConfigSpec(glob_path, config_type=".json", check_if_exists=False),
                os.environ,
            ]
        )
        return ClientConfig.from_dict(config.data)



from typing import List, Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.lifecycle import V1StatusCondition
from polyaxon.schemas.base import BaseResponseModel


class V1EntityNotificationBody(BaseResponseModel):
    namespace: Optional[StrictStr]
    owner: Optional[StrictStr]
    project: Optional[StrictStr]
    uuid: Optional[UUIDStr]
    name: Optional[StrictStr]
    condition: Optional[V1StatusCondition]
    connections: Optional[List[StrictStr]]



from clipped.config.manager import ConfigManager as _ConfigManager

from polyaxon.config.reader import ConfigReader
from polyaxon.contexts import paths as ctx_paths
from polyaxon.logger import logger


class ConfigManager(_ConfigManager):
    _CONFIG_READER = ConfigReader
    _LOGGER = logger
    _PROJECT = ".polyaxon"
    _PROJECT_PATH: str = ctx_paths.CONTEXT_USER_POLYAXON_PATH
    _TEMP_PATH: str = ctx_paths.CONTEXT_TMP_POLYAXON_PATH



import uuid

from typing import Optional

from polyaxon.contexts.params import PARAM_REGEX, is_template_ref
from polyaxon.exceptions import PolyaxonValidationError

OPS = "ops"
RUNS = "runs"
DAG = "dag"
DAG_ENTITY_REF = "_"
ENTITIES = {OPS, RUNS}

ENTITY_REF_FORMAT = "{}.{}"


def is_runs_ref(ref: str) -> bool:
    return ref.split(".")[0] == RUNS


def is_ops_ref(ref: str) -> bool:
    return ref.split(".")[0] == OPS


def is_dag_ref(ref: str) -> bool:
    return ref.split(".")[0] == DAG


def get_entity_ref(ref: str) -> Optional[str]:
    if is_ops_ref(ref) or is_runs_ref(ref):
        return ref.split(".")[1]
    if is_dag_ref(ref):
        return DAG_ENTITY_REF
    return None


def get_entity_type(value: str) -> str:
    value_parts = PARAM_REGEX.search(value)
    if value_parts:
        results = value_parts.group(1).strip()
    else:
        results = value

    return results.split(".")[0]


def get_entity_value(value: str) -> Optional[str]:
    value_parts = PARAM_REGEX.search(value)
    if value_parts:
        results = value_parts.group(1).strip()
    else:
        results = value

    results = results.split(".")
    if len(results) < 2:
        return None
    return results[-1]


def parse_ref_value(value: str) -> str:
    """Returns value without {{ }}"""
    value_parts = PARAM_REGEX.search(value)
    if value_parts:
        return value_parts.group(1).strip()
    return value


class RefMixin:
    @property
    def is_literal(self) -> bool:
        return not self.ref  # type: ignore[attr-defined]

    @property
    def is_ref(self) -> bool:
        return self.ref is not None  # type: ignore[attr-defined]

    @property
    def is_template_ref(self) -> bool:
        return is_template_ref(self.value)

    @property
    def is_runs_ref(self) -> bool:
        if not self.is_ref:
            return False

        return is_runs_ref(self.ref)  # type: ignore[attr-defined]

    @property
    def is_ops_ref(self) -> bool:
        if not self.is_ref:
            return False

        return is_ops_ref(self.ref)  # type: ignore[attr-defined]

    @property
    def is_dag_ref(self) -> bool:
        if not self.is_ref:
            return False

        return is_dag_ref(self.ref)  # type: ignore[attr-defined]

    @property
    def is_join_ref(self):
        return False

    @property
    def entity_ref(self) -> Optional[str]:
        return get_entity_ref(self.ref)  # type: ignore[attr-defined]


def validate_ref(ref: str, name: str):
    # validate ref
    ref_parts = ref.split(".")
    if len(ref_parts) > 2:
        raise PolyaxonValidationError(
            "Could not parse ref `{}` for param `{}`.".format(ref, name)
        )
    if len(ref_parts) == 1 and ref_parts[0] != DAG:
        raise PolyaxonValidationError(
            "Could not parse ref `{}` for param `{}`.".format(ref_parts[0], name)
        )
    if len(ref_parts) == 2 and ref_parts[0] not in ENTITIES:
        raise PolyaxonValidationError(
            "Could not parse ref `{}` for param `{}`. "
            "Operation ref must be one of `{}`".format(ref_parts[0], name, ENTITIES)
        )
    if ref_parts[0] == RUNS:
        try:
            uuid.UUID(ref_parts[1])
        except (KeyError, ValueError):
            raise PolyaxonValidationError(
                "Param value `{}` should reference a valid run uuid.".format(
                    ref_parts[1]
                )
            )



from typing import Any, List, Optional

from clipped.config.schema import skip_partial
from pydantic import Field, StrictStr, root_validator, validator

from polyaxon.config.parser import ConfigParser
from polyaxon.exceptions import PolyaxonSchemaError, PolyaxonValidationError
from polyaxon.schemas.base import BaseSchemaModel

IO_NAME_BLACK_LIST = ["globals", "params", "connections"]
IO_NAME_ERROR = (
    "Received an input/output with a name in {}, "
    "please use a different name that does "
    "not already taken by the context".format(IO_NAME_BLACK_LIST)
)


def validate_io_value(
    name: str,
    type: str,
    value: Any,
    default: Any,
    is_optional: bool,
    is_list: bool,
    options: List[Any],
    parse: bool = True,
):
    try:
        parsed_value = ConfigParser.parse(type)(
            key=name,
            value=value,
            is_list=is_list,
            is_optional=is_optional,
            default=default,
            options=options,
        )
        if parse:
            return parsed_value
        # Return the original value, the parser will return specs sometimes
        if value is not None:
            return value
        return default
    except PolyaxonSchemaError as e:
        raise PolyaxonValidationError(
            "Could not parse value `%s`, an error was encountered: %s" % (value, e)
        )


def validate_io(name, type, value, is_optional, is_list, is_flag, options):
    if type and value:
        try:
            value = validate_io_value(
                name=name,
                type=type,
                value=value,
                default=None,
                is_list=is_list,
                is_optional=is_optional,
                options=options,
            )
        except PolyaxonValidationError as e:
            raise ValueError(e)

    if not is_optional and value:
        raise ValueError(
            "IO `{}` is not optional and has default value `{}`. "
            "Please either make it optional or remove the default value.".format(
                name, value
            )
        )

    if is_flag and type != "bool":
        raise TypeError(
            "IO type `{}` cannot be a flag, it must be of type `{}`".format(
                type, "bool"
            )
        )

    return value


class V1IO(BaseSchemaModel):
    """Each Component may have its own inputs and outputs.
    The inputs and outputs describe the expected parameters to pass to the component
    and their types. In the context of a DAG,
    inputs and outputs types are used to validate the flow of information
    going from one operation to another.

    The final value of an input/output can be resolved
    from [params](/docs/core/specification/params/), or from other values in
    the [context](/docs/core/context/).

    Examples:
     * A build component may have a git repository as input and a container image as output.
     * A training component may have a container image, data path,
       and some hyperparameters as input and a list of metrics and artifacts as outputs.

    An input/output section includes a name, a description, a type to check the value passed,
    a flag to tell if the input/output is optional, and a default value if it's optional.

    Sometimes users prefer to pass a param to the `command` or `args`
    section only if the value is not null. Polyaxon provides a way to do that using
    `{{ params.PARAM_NAME.as_arg }}`,
    this will be empty if the value is null or `--PARAM_NAME=value` if the value is not null.
    If the value is of type bool, it will be `--PARAM_NAME`.
    It's also possible to control how the param should be
    converted to an argument using the `arg_format`.

    To learn how to pass valid values,
    please check the [param section](/docs/core/specification/params/).

    Args:
        name: str
        description: str, optional
        type: str, any python type hint, pydantic built-in types, and gcs, s3, wasb, dockerfile, git, image, event, artifacts, path, metric, metadata, date, datetime.
        value: any, optional
        is_optional: bool, optional (**Deprecated**)
        is_list: bool, optional
        is_flag: bool, optional
        arg_format: str, optional
        delay_validation: bool, optional
        options: List[any], optional (**Deprecated**)
        validation: Validation, optional
        connection: str, optional
        to_init: bool, optional
        to_env: str, optional

    ## YAML usage

    ```yaml
    >>> inputs:
    >>>   - name: loss
    >>>     type: str
    >>>     isOptional: true
    >>>     value: MeanSquaredError
    >>>   - name: preprocess
    >>>     type: bool
    >>> outputs:
    >>>   - name: accuracy
    >>>     type: float
    >>>   - name: outputs-path
    >>>     type: path
    ```

    ## Python usage

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.polyflow import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="loss",
    >>>         type='str',
    >>>         description="Loss to use for training my model",
    >>>         is_optional=True,
    >>>         value="MeanSquaredError"
    >>>     ),
    >>>     V1IO(
    >>>         name="preprocess",
    >>>         type='bool',
    >>>         description="A flag to preprocess data before training",
    >>>     )
    >>> ]
    >>> outputs = [
    >>>     V1IO(
    >>>         name="accuracy",
    >>>         type='float',
    >>>     ),
    >>>     V1IO(
    >>>         name="outputs-path",
    >>>         type=types.PATH,
    >>>     )
    >>> ]
    ```

    These inputs/outputs declarations can be used to pass values to our program:

    ```bash
     ... --loss={{ loss }} {{ params.preprocess.as_arg }}
    ```

    ## Fields

    ### name

    The input/output name. The name must be a valid slug, and cannot include dots `.`.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    ```

    ### description

    An optional description for the input/output.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    ```

    ### type

    The type of the input/output. The type will be used to validate the value

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    ```

    for more details about composite type validation and schema,
    please check the [types section](/docs/core/specification/types/),
    possible types include any python type hint, pydantic built-in types, and:
        * URI: "uri"
        * LIST: "list"
        * GCS: "gcs"
        * S3: "s3"
        * WASB: "wasb"
        * DOCKERFILE: "dockerfile"
        * GIT: "git"
        * IMAGE: "image"
        * EVENT: "event"
        * ARTIFACTS: "artifacts"
        * PATH: "path"
        * METRIC: "metric"
        * METADATA: "metadata"

    ### value

    If an input is optional you should assign it a value.
    If an output is optional you can assign it a value.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 1.1
    ```

    ### isOptional

    A flag to tell if an input/output is optional.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 1.1
    >>>     isOptional: true
    ```

    ### isList

    > **Deprecated**: Please use `type: List[TYPING]` instead.

    In `v2` you should:

    ```yaml
    >>> inputs:
    >>>   - name: learning_rates
    >>>     type: List[float]
    ```

    A flag used in `v1` to tell if an input/output is a list of the type passed.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rates
    >>>     type: float
    >>>     isList: true
    ```

    In this case the input name `learning_rates` will expect a value of type `List[float]`,
    e.g. [0.1 0.01, 0.0001]

    ### argFormat

    A key to control how to convert an input/output to a CLI argument if the value is not null,
    the default behavior:

     * For bool type: If the resolved value of the input is True,
       `"{{ params.PARAM_NAME.as_arg }}"` will be resolved to `"--PARAM_NAME"`
       otherwise it will be an empty string `""`.
     * For non-bool types: If the resolved value of the input is not null,
       `"{{ params.PARAM_NAME.as_arg }}"` will be resolved to `"--PARAM_NAME=value"`
       otherwise it will be an empty string `""`.

    Let's look at the flowing example:

    ```yaml
    >>> inputs:
    >>>   - name: lr
    >>>     type: float
    >>>   - name: check
    >>>     type: bool
    ```

    **This manifest**:

    ```yaml
    >>> container:
    >>>    command: ["{{ lr }}", "{{ check }}"]
    ```

    Will be transformed to:

    ```yaml
    >>> container:
    >>>    command: ["0.01", "true"]
    ```

    **This manifest**:

    ```yaml
    >>> container:
    >>>    command: ["{{ params.lr.as_arg }}", "{{ params.check.as_arg }}"]
    ```

    Will be transformed to:

    ```yaml
    >>> container:
    >>>    command: ["--lr=0.01", "--check"]
    ```

    Changing the behavior with `argFormat`:

    ```yaml
    >>> inputs:
    >>>   - name: lr
    >>>     type: float
    >>>     argFormat: "lr={{ lr }}"
    >>>   - name: check
    >>>     type: bool
    >>>     argFormat: "{{ 1 if check else 0 }}"
    ```

    **This manifest**:

    ```yaml
    >>> container:
    >>>    command: ["{{ params.lr.as_arg }}", "{{ params.check.as_arg }}"]
    ```

    Will be transformed to:

    ```yaml
    >>> container:
    >>>    command: ["lr=0.01", "1"]
    ```

    ### delayValidation

    A flag to tell if an input/output should not be
    validated at compilation or resolution time.

    This flag is enabled by default for outputs, since they can only be
    resolved after or during the run. To request validation at compilation time for outputs,
    you need to set this flag to `False`.

    ### connection

    A connection to use with the input/outputs.

    ### toInit

    If True, it will be converted to an init container.

    ### toEnv

    > **N.B**: Requires Polyaxon CLI and Polyaxon Agent/CE version `>= 1.12`

    If passed, it will be converted automatically to an environment variable.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     type: float
    >>>     value: 1.1
    >>>     toEnv: MY_LEARNING_RATE
    ```

    ### options

    > **Deprecated**: Please use `validation: ...` instead.

    Options allow to pass a list of values that will be used to validate any passed params.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 1.1
    >>>     options: [1.1, 2.2, 3.3]
    ```

    If you pass the value `4.4` for the learning rate it will raise a validation error.

    ## Example


    ```yaml
    >>> version: 1.1
    >>> kind: component
    >>> inputs:
    >>>   - name: batch_size
    >>>     description: batch size
    >>>     isOptional: true
    >>>     value: 128
    >>>     type: int
    >>>   - name: num_steps
    >>>     isOptional: true
    >>>     default: 500
    >>>     type: int
    >>>   - name: learning_rate
    >>>     isOptional: true
    >>>     default: 0.001
    >>>     type: float
    >>>   - name: dropout
    >>>     isOptional: true
    >>>     default: 0.25
    >>>     type: float
    >>>   - name: num_epochs
    >>>     isOptional: true
    >>>     default: 1
    >>>     type: int
    >>>   - name: activation
    >>>     isOptional: true
    >>>     default: relu
    >>>     type: str
    >>> run:
    >>>   kind: job
    >>>   image: foo:bar
    >>>   container:
    >>>     command: [python3, model.py]
    >>>     args: [
    >>>         "--batch_size={{ batch_size }}",
    >>>         "--num_steps={{ num_steps }}",
    >>>         "--learning_rate={{ learning_rate }}",
    >>>         "--dropout={{ dropout }",
    >>>         "--num_epochs={{ num_epochs }}",
    >>>         "--activation={{ activation }}"
    >>>     ]
    ```

    ### Running a typed component using the CLI

    Using the Polyaxon CLI we can now run this component and override the inputs' default values:

    ```bash
    polyaxon run -f polyaxonfile.yaml -P activation=sigmoid -P dropout=0.4
    ```

    This will generate a manifest and will replace the params passed and validated against the inputs' types.

    ### Required inputs

    In the example all inputs are optional.
    If we decide for instance to make the activation required:

    ````yaml
    >>> ...
    >>> inputs:
    >>>   ...
    >>>   - name: activation
    >>>     type: str
    >>>   ...
    ...
    ````

    By changing this input, polyaxon can not run this component without passing the activation:


    ```bash
    polyaxon run -f polyaxonfile.yaml -P activation=sigmoid
    ```
    """

    _IDENTIFIER = "io"

    name: StrictStr
    description: Optional[StrictStr]
    type: Optional[StrictStr]
    is_optional: Optional[bool] = Field(alias="isOptional")
    is_list: Optional[bool] = Field(alias="isList")
    is_flag: Optional[bool] = Field(alias="isFlag")
    arg_format: Optional[StrictStr] = Field(alias="argFormat")
    delay_validation: Optional[bool] = Field(alias="delayValidation")
    options: Optional[Any]
    connection: Optional[StrictStr]
    to_init: Optional[bool] = Field(alias="toInit")
    to_env: Optional[StrictStr] = Field(alias="toEnv")
    value: Optional[Any]

    @validator("name", always=True)
    def validate_name(cls, v):
        if v in IO_NAME_BLACK_LIST:
            raise ValueError(IO_NAME_ERROR)
        return v

    @root_validator
    @skip_partial
    def validate_io(cls, values):
        validate_io(
            name=values.get("name"),
            type=values.get("type"),
            value=values.get("value"),
            is_list=values.get("is_list"),
            is_optional=values.get("is_optional"),
            is_flag=values.get("is_flag"),
            options=values.get("options"),
        )
        return values

    def validate_value(self, value: Any, parse: bool = True):
        if self.type is None:
            return value

        return validate_io_value(
            name=self.name,
            type=self.type,
            value=value,
            default=self.value,
            is_list=self.is_list,
            is_optional=self.is_optional,
            options=self.options,
            parse=parse,
        )

    def get_repr_from_value(self, value):
        """A string representation that is used to create hash cache"""
        value = self.validate_value(value=value, parse=False)
        io_dict = self.to_light_dict(include_attrs=["name", "type"])
        io_dict["value"] = value
        return io_dict

    def get_repr(self):
        """A string representation that is used to create hash cache"""
        io_dict = self.to_light_dict(include_attrs=["name", "type", "value"])
        return io_dict



from typing import Type

from clipped.formatting import Printer

from polyaxon.config.manager import ConfigManager
from polyaxon.schemas.responses.v1_user import V1User


class UserConfigManager(ConfigManager):
    """Manages user configuration .user file."""

    VISIBILITY = ConfigManager.Visibility.GLOBAL
    IN_PROJECT_DIR = True
    CONFIG_FILE_NAME = ".user"
    CONFIG: Type[V1User] = V1User

    @classmethod
    def get_config_or_raise(cls) -> V1User:
        user = cls.get_config()
        if not user:
            Printer.error("User configuration was not found.", sys_exit=True)

        return user

    @classmethod
    def get_config_from_env(cls):
        pass



from typing import List, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_tag import V1Tag


class V1ListTagsResponse(BaseResponseModel):
    count: Optional[int]
    results: Optional[List[V1Tag]]
    previous: Optional[StrictStr]
    next: Optional[StrictStr]



from polyaxon.polyflow import V1RunKind


def get_fxt_job():
    return {
        "version": 1.1,
        "kind": "operation",
        "name": "foo",
        "description": "a description",
        "tags": ["tag1", "tag2"],
        "trigger": "all_succeeded",
        "component": {
            "name": "build-template",
            "tags": ["tag1", "tag2"],
            "run": {
                "kind": V1RunKind.JOB,
                "container": {"image": "test"},
                "init": [{"connection": "foo", "git": {"revision": "dev"}}],
            },
        },
    }


def get_fxt_job_with_inputs():
    return {
        "version": 1.1,
        "kind": "operation",
        "name": "foo",
        "description": "a description",
        "params": {"image": {"value": "foo/bar"}},
        "component": {
            "name": "build-template",
            "inputs": [{"name": "image", "type": "str"}],
            "tags": ["tag1", "tag2"],
            "run": {
                "kind": V1RunKind.JOB,
                "container": {
                    "image": "{{ image }}",
                    "command": ["foo"],
                    "args": ["foo"],
                },
                "init": [{"connection": "foo", "git": {"revision": "dev"}}],
            },
        },
    }


def get_fxt_job_with_inputs_outputs():
    return {
        "version": 1.1,
        "kind": "operation",
        "name": "foo",
        "description": "a description",
        "cost": 2.2,
        "params": {"image": {"value": "foo/bar"}},
        "component": {
            "name": "build-template",
            "inputs": [{"name": "image", "type": "str"}],
            "outputs": [
                {"name": "result1", "type": "str"},
                {
                    "name": "result2",
                    "type": "str",
                    "isOptional": True,
                    "value": "{{ image }}",
                },
            ],
            "tags": ["tag1", "tag2"],
            "run": {"kind": V1RunKind.JOB, "container": {"image": "{{ image }}"}},
        },
    }


def get_fxt_job_with_inputs_and_conditions():
    return {
        "version": 1.1,
        "kind": "operation",
        "name": "foo",
        "description": "a description",
        "cost": 0.3,
        "params": {
            "image": {"value": "foo/bar"},
            "context_param": {"value": "some-value", "contextOnly": True},
            "init_param": {
                "value": {"url": "https://git.url"},
                "connection": "git2",
                "toInit": True,
            },
        },
        "conditions": "{{ image == 'foo/bar' and context_param == 'some-value' }}",
        "component": {
            "name": "build-template",
            "inputs": [
                {"name": "image", "type": "str"},
                {"name": "init_param", "type": "git"},
            ],
            "tags": ["tag1", "tag2"],
            "run": {
                "kind": V1RunKind.JOB,
                "container": {"image": "{{ image }}"},
                "init": [{"connection": "foo", "git": {"revision": "dev"}}],
            },
        },
    }


def get_fxt_job_with_inputs_and_joins():
    return {
        "version": 1.1,
        "kind": "operation",
        "name": "foo",
        "description": "a description",
        "joins": [
            {
                "query": "metrics.metric: > {{ metric_level }}",
                "sort": "-metrics.metric",
                "limit": "{{ top }}",
                "params": {
                    "metrics": {"value": "outputs.metric"},
                },
            },
            {
                "query": "metrics.metric: < -1",
                "params": {
                    "excluded": {"value": "outputs.metric", "contextOnly": True},
                    "excluded_statuses": {
                        "value": "globals.status",
                        "contextOnly": True,
                    },
                    "excluded_uuids": {"value": "globals.uuid"},
                    "run_metrics_events": {
                        "value": "artifacts.metric1",
                        "contextOnly": True,
                    },
                    "artifact_paths": {
                        "value": "artifacts.base",
                    },
                    "run_outputs_paths": {
                        "value": "artifacts.outputs",
                    },
                    "all_outputs": {
                        "value": "outputs",
                    },
                    "all_inputs": {
                        "value": "inputs",
                    },
                    "all_artifacts": {
                        "value": "artifacts",
                    },
                    "files": {
                        "value": {
                            "files": ["subpath/file1", "different/subpath/file2"],
                        },
                        "toInit": True,
                    },
                    "files2": {
                        "value": {
                            "files": ["subpath/file21", "different/subpath/file22"],
                        },
                        "contextOnly": True,
                    },
                },
            },
        ],
        "params": {
            "top": {"value": 3, "contextOnly": True},
            "metric_level": {"value": 0, "contextOnly": True},
            "init_param": {
                "value": {"url": "https://git.url"},
                "connection": "git2",
                "toInit": True,
            },
        },
        "component": {
            "name": "build-template",
            "inputs": [
                {"name": "metrics", "type": "float", "isList": True},
                {"name": "excluded_uuids", "type": "uuid", "isList": True},
                {"name": "init_param", "type": "git"},
                {"name": "files", "type": "artifacts"},
                {"name": "artifact_paths", "type": "path", "isList": True},
                {"name": "run_outputs_paths", "type": "path", "isList": True},
                {"name": "all_outputs", "type": "dict", "isList": True},
                {"name": "all_inputs", "type": "dict", "isList": True},
                {"name": "all_artifacts", "type": "dict", "isList": True},
            ],
            "tags": ["tag1", "tag2"],
            "run": {
                "kind": V1RunKind.JOB,
                "container": {"image": "test"},
                "init": [
                    {"connection": "foo", "git": {"revision": "dev"}},
                    {"artifacts": {"dirs": "{{run_outputs}}"}},
                ],
            },
        },
    }


def get_fxt_tf_job():
    return {
        "version": 1.1,
        "kind": "operation",
        "name": "foo",
        "description": "a description",
        "component": {
            "name": "tf-distributed-gpu",
            "tags": ["tag1", "tag2"],
            "run": {
                "kind": "tfjob",
                "chief": {
                    "environment": {
                        "nodeSelector": {"polyaxon": "experiments-gpu-t4"},
                        "tolerations": [
                            {
                                "key": "nvidia.com/gpu",
                                "operator": "Equal",
                                "value": "present",
                                "effect": "NoSchedule",
                            }
                        ],
                    },
                    "container": {
                        "resources": {
                            "requests": {"cpu": 4, "memory": "4Gi"},
                            "limits": {"nvidia.com/gpu": 1, "cpu": 4, "memory": "8Gi"},
                        },
                        "image": "foo/bar:gpu",
                        "workingDir": "{{ globals.run_artifacts_path }}/uploads/src",
                        "command": ["python", "-u", "mnist.py"],
                    },
                },
                "worker": {
                    "replicas": 2,
                    "environment": {
                        "restartPolicy": "OnFailure",
                        "nodeSelector": {"polyaxon": "experiments-gpu-t4"},
                        "tolerations": [
                            {
                                "key": "nvidia.com/gpu",
                                "operator": "Equal",
                                "value": "present",
                                "effect": "NoSchedule",
                            }
                        ],
                    },
                    "container": {
                        "resources": {
                            "requests": {"cpu": 4, "memory": "4Gi"},
                            "limits": {"nvidia.com/gpu": 1, "cpu": 4, "memory": "8Gi"},
                        },
                        "image": "foo/bar:gpu",
                        "workingDir": "{{ globals.run_artifacts_path }}/uploads/src",
                        "command": ["python", "-u", "mnist.py"],
                    },
                },
            },
        },
    }


def get_fxt_ray_job():
    return {
        "version": 1.1,
        "kind": "operation",
        "name": "foo",
        "description": "a description",
        "component": {
            "name": "ray-job",
            "tags": ["tag1", "tag2"],
            "run": {
                "kind": "rayjob",
                "entrypoint": "python sample_code.py",
                "runtimeEnv": {
                    "pip": ["requests==2.26.0", "pendulum==2.1.2"],
                    "env_vars": {"counter_name": "test_counter"},
                },
                "rayVersion": "2.5.0",
                "head": {
                    "environment": {
                        "nodeSelector": {"polyaxon": "experiments-gpu-t4"},
                        "tolerations": [
                            {
                                "key": "nvidia.com/gpu",
                                "operator": "Equal",
                                "value": "present",
                                "effect": "NoSchedule",
                            }
                        ],
                    },
                    "container": {
                        "resources": {
                            "requests": {"cpu": 4, "memory": "4Gi"},
                            "limits": {"nvidia.com/gpu": 1, "cpu": 4, "memory": "8Gi"},
                        },
                        "image": "foo/bar:gpu",
                        "workingDir": "{{ globals.run_artifacts_path }}/uploads/src",
                        "command": ["python", "-u", "mnist.py"],
                    },
                },
                "workers": {
                    "small-group": {
                        "replicas": 2,
                        "minReplicas": 2,
                        "maxReplicas": 4,
                        "rayStartParams": {},
                        "environment": {
                            "restartPolicy": "OnFailure",
                            "nodeSelector": {"polyaxon": "experiments-gpu-t4"},
                            "tolerations": [
                                {
                                    "key": "nvidia.com/gpu",
                                    "operator": "Equal",
                                    "value": "present",
                                    "effect": "NoSchedule",
                                }
                            ],
                        },
                        "container": {
                            "resources": {
                                "requests": {"cpu": 4, "memory": "4Gi"},
                                "limits": {
                                    "nvidia.com/gpu": 1,
                                    "cpu": 4,
                                    "memory": "8Gi",
                                },
                            },
                            "image": "foo/bar:gpu",
                            "lifecycle": {
                                "preStop": {
                                    "exec": {"command": ["/bin/sh", "-c", "ray stop"]}
                                }
                            },
                        },
                    }
                },
            },
        },
    }



from typing import Type

from polyaxon.config.manager import ConfigManager
from polyaxon.polyflow import V1Init
from polyaxon.schemas.types import V1GitType


class GitConfigManager(ConfigManager):
    """Manages access token configuration .auth file."""

    VISIBILITY = ConfigManager.Visibility.LOCAL
    CONFIG_FILE_NAME = "polyaxongit.yaml"
    CONFIG: Type[V1Init] = V1Init

    @classmethod
    def get_config_from_env(cls) -> V1Init:
        pass

    @classmethod
    def get_config(cls, check: bool = True) -> V1Init:
        config = super(GitConfigManager, cls).get_config(check=check)
        if not config.git:
            config.git = V1GitType()

        return config



from typing import Dict, Optional

from clipped.formatting import Printer

from polyaxon.exceptions import handle_api_error


def handle_cli_error(
    e,
    message: Optional[str] = None,
    http_messages_mapping: Optional[Dict] = None,
    sys_exit: bool = False,
):
    handle_api_error(
        logger=Printer,
        e=e,
        message=message,
        http_messages_mapping=http_messages_mapping,
        sys_exit=sys_exit,
    )


def handle_command_not_in_ce():
    Printer.error(
        "You are running Polyaxon CE which does not support this command!",
        sys_exit=True,
    )



from polyaxon.k8s.k8s_schemas import V1Container


class DestinationImageMixin:
    def apply_image_destination(self, image: str):
        self.container = self.container or V1Container()
        self.container.image = image



import atexit
import re

from multiprocessing.pool import ThreadPool
from urllib.parse import quote

from polyaxon.sdk.exceptions import ApiException
from polyaxon.sdk.sync_client.api_client import ApiClient


class AsyncApiClient(ApiClient):
    """Generic Async API client for OpenAPI client library builds.

    OpenAPI generic API client. This client handles the client-
    server communication, and is invariant across implementations. Specifics of
    the methods and models for each application are generated from the OpenAPI
    templates.

    :param configuration: .Configuration object for this client
    :param header_name: a header to pass when making calls to the API.
    :param header_value: a header value to pass when making calls to
        the API.
    :param cookie: a cookie to include in the header when making calls
        to the API
    :param pool_threads: The number of threads to use for async requests
        to the API. More threads means more concurrent API requests.
    """

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()

    async def close(self):
        await self.rest_client.close()
        if self._pool:
            self._pool.close()
            self._pool.join()
            self._pool = None
            if hasattr(atexit, "unregister"):
                atexit.unregister(self.close)

    def _get_rest_client(self):
        from polyaxon.sdk.async_client import rest

        return rest.RESTClientObject(self.configuration)

    @property
    def pool(self):
        """Create thread pool on first request
        avoids instantiating unused threadpool for blocking clients.
        """
        if self._pool is None:
            atexit.register(self.close)
            self._pool = ThreadPool(self.pool_threads)
        return self._pool

    @property
    def user_agent(self):
        """User agent for this API client"""
        return self.default_headers["User-Agent"]

    @user_agent.setter
    def user_agent(self, value):
        self.default_headers["User-Agent"] = value

    def set_default_header(self, header_name, header_value):
        self.default_headers[header_name] = header_value

    _default = None

    @classmethod
    def set_default(cls, default):
        """Set default instance of AsyncApiClient.

        It stores default AsyncApiClient.

        :param default: object of AsyncApiClient.
        """
        cls._default = default

    async def __call_api(
        self,
        resource_path,
        method,
        path_params=None,
        query_params=None,
        header_params=None,
        body=None,
        post_params=None,
        files=None,
        response_types_map=None,
        auth_settings=None,
        _return_http_data_only=None,
        collection_formats=None,
        _preload_content=True,
        _request_timeout=None,
        _host=None,
        _request_auth=None,
    ):
        config = self.configuration

        # header parameters
        header_params = header_params or {}
        header_params.update(self.default_headers)
        if self.cookie:
            header_params["Cookie"] = self.cookie
        if header_params:
            header_params = self.sanitize_for_serialization(header_params)
            header_params = dict(
                self.parameters_to_tuples(header_params, collection_formats)
            )

        # path parameters
        if path_params:
            path_params = self.sanitize_for_serialization(path_params)
            path_params = self.parameters_to_tuples(path_params, collection_formats)
            for k, v in path_params:
                # specified safe chars, encode everything
                resource_path = resource_path.replace(
                    "{%s}" % k, quote(str(v), safe=config.safe_chars_for_path_param)
                )

        # post parameters
        if post_params or files:
            post_params = post_params if post_params else []
            post_params = self.sanitize_for_serialization(post_params)
            post_params = self.parameters_to_tuples(post_params, collection_formats)
            post_params.extend(self.files_parameters(files))

        # auth setting
        self.update_params_for_auth(
            header_params,
            query_params,
            auth_settings,
            resource_path,
            method,
            body,
            request_auth=_request_auth,
        )

        # body
        if body:
            body = self.sanitize_for_serialization(body)

        # request url
        if _host is None:
            url = self.configuration.host + resource_path
        else:
            # use server/host defined in path or operation instead
            url = _host + resource_path

        # query parameters
        if query_params:
            query_params = self.sanitize_for_serialization(query_params)
            url_query = self.parameters_to_url_query(query_params, collection_formats)
            url += "?" + url_query

        try:
            # perform request and return response
            response_data = await self.request(
                method,
                url,
                query_params=query_params,
                headers=header_params,
                post_params=post_params,
                body=body,
                _preload_content=_preload_content,
                _request_timeout=_request_timeout,
            )
        except ApiException as e:
            if e.body:
                e.body = e.body.decode("utf-8")
            raise e

        self.last_response = response_data

        return_data = response_data

        if not _preload_content:
            return return_data

        response_type = response_types_map.get(str(response_data.status), None)

        if response_type == "bytearray":
            response_data.data = response_data.data
        else:
            match = None
            content_type = response_data.getheader("content-type")
            if content_type is not None:
                match = re.search(r"charset=([a-zA-Z\-\d]+)[\s;]?", content_type)
            encoding = match.group(1) if match else "utf-8"
            response_data.data = response_data.data.decode(encoding)

            # deserialize response data
            if response_type == "bytearray":
                return_data = response_data.data
            elif response_type:
                return_data = self.deserialize(response_data, response_type)
            else:
                return_data = None

        if _return_http_data_only:
            return return_data
        else:
            return (return_data, response_data.status, response_data.getheaders())



from vents.config import AppConfig
from vents.settings import create_app

from polyaxon.config.parser import ConfigParser
from polyaxon.contexts import paths as ctx_paths
from polyaxon.exceptions import PolyaxonConnectionError
from polyaxon.logger import logger

CONNECTION_CONFIG = create_app(
    config=AppConfig(
        project_name="Polyaxon",
        project_url="https://polyaxon.com/",
        project_icon="https://cdn.polyaxon.com/static/v1/images/logo_small.png",
        env_prefix="POLYAXON",
        context_path=ctx_paths.CONTEXT_ROOT,
        logger=logger,
        exception=PolyaxonConnectionError,
        config_parser=ConfigParser,
    )
)

from polyaxon.connections.kinds import V1ConnectionKind
from polyaxon.connections.schemas import (
    V1BucketConnection,
    V1ClaimConnection,
    V1Connection,
    V1ConnectionResource,
    V1GitConnection,
    V1HostConnection,
    V1HostPathConnection,
)



import atexit
import sys
import time
import traceback

from concurrent.futures import ThreadPoolExecutor
from typing import Any, Dict, Optional, Tuple, Type

from clipped.utils.tz import now
from clipped.utils.versions import clean_version_for_check
from clipped.utils.workers import exit_context, get_pool_workers, get_wait
from kubernetes.client.rest import ApiException
from urllib3.exceptions import HTTPError

from polyaxon import pkg, settings
from polyaxon.auxiliaries import V1PolyaxonInitContainer, V1PolyaxonSidecarContainer
from polyaxon.client import PolyaxonClient
from polyaxon.connections import V1Connection
from polyaxon.env_vars.getters import get_run_info
from polyaxon.exceptions import PolyaxonConverterError
from polyaxon.lifecycle import LiveState, V1StatusCondition, V1Statuses
from polyaxon.logger import logger
from polyaxon.runner.executor import BaseExecutor
from polyaxon.schemas.cli.checks_config import ChecksConfig
from polyaxon.schemas.responses.v1_agent import V1Agent
from polyaxon.schemas.responses.v1_agent_state_response import V1AgentStateResponse
from polyaxon.sdk.exceptions import ApiException


class BaseAgent:
    EXECUTOR: Type[BaseExecutor]
    HEALTH_FILE = "/tmp/.healthz"
    SLEEP_STOP_TIME = 60 * 5
    SLEEP_ARCHIVED_TIME = 60 * 60

    def __init__(
        self,
        owner: Optional[str] = None,
        agent_uuid: Optional[str] = None,
        sleep_interval: Optional[int] = None,
    ):
        self.owner = owner
        self.agent_uuid = agent_uuid
        self.sleep_interval = sleep_interval
        if self.sleep_interval:
            self.sleep_interval = max(self.sleep_interval, 1)
        self.executor = None
        self._executor_refreshed_at = now()
        self._graceful_shutdown = False
        self.client = PolyaxonClient()
        self.executor = self.EXECUTOR()
        self.content = settings.AGENT_CONFIG.to_json()
        self._register()

    def _is_managed(self):
        return self.owner and self.agent_uuid

    def _register(self):
        if not self._is_managed:
            return
        print("Agent is starting.")
        try:
            agent = self.get_info()
            self._check_status(agent)
            self.sync()
            self.log_agent_running()
            print("Agent is running.")
        except (ApiException, HTTPError) as e:
            self.log_agent_failed(
                message="Could not start the agent {}.".format(repr(e))
            )
            sys.exit(1)
        atexit.register(self._wait)

    def _wait(self):
        if not self._is_managed:
            return
        if not self._graceful_shutdown:
            self.log_agent_warning()
        time.sleep(1)

    def get_info(self) -> V1Agent:
        return self.client.agents_v1.get_agent(owner=self.owner, uuid=self.agent_uuid)

    def get_state(self) -> V1AgentStateResponse:
        if self._is_managed:
            return self.client.agents_v1.get_agent_state(
                owner=self.owner, uuid=self.agent_uuid
            )
        return self.client.agents_v1.get_global_state(owner=self.owner)

    def log_agent_status(
        self, status: str, reason: Optional[str] = None, message: Optional[str] = None
    ):
        if not self._is_managed:
            return
        status_condition = V1StatusCondition.get_condition(
            type=status, status=True, reason=reason, message=message
        )
        self.client.agents_v1.create_agent_status(
            owner=self.owner,
            uuid=self.agent_uuid,
            body={"condition": status_condition},
            async_req=True,
        )

    def sync(self):
        self.client.agents_v1.sync_agent(
            owner=self.owner,
            agent_uuid=self.agent_uuid,
            body=V1Agent(
                content=settings.AGENT_CONFIG.to_json(),
                version=clean_version_for_check(pkg.VERSION),
                version_api=self.executor.manager.get_version(),
            ),
        )

    def sync_compatible_updates(self, compatible_updates: Dict):
        if compatible_updates and settings.AGENT_CONFIG:
            init = compatible_updates.get("init")
            if init and settings.AGENT_CONFIG.init:
                init = V1PolyaxonInitContainer.from_dict(init)
                settings.AGENT_CONFIG.init = settings.AGENT_CONFIG.init.patch(init)

            sidecar = compatible_updates.get("sidecar")
            if sidecar and settings.AGENT_CONFIG.sidecar:
                sidecar = V1PolyaxonSidecarContainer.from_dict(sidecar)
                settings.AGENT_CONFIG.sidecar = settings.AGENT_CONFIG.sidecar.patch(
                    sidecar
                )
            connections = compatible_updates.get("connections")
            if connections:
                settings.AGENT_CONFIG.connections = [
                    V1Connection.from_dict(c) for c in connections
                ]

            self.content = settings.AGENT_CONFIG.to_json()
            self.sync()

    def log_agent_running(self):
        self.log_agent_status(status=V1Statuses.RUNNING, reason="AgentLogger")

    def log_agent_failed(self, message=None):
        self.log_agent_status(
            status=V1Statuses.FAILED, reason="AgentLogger", message=message
        )

    def log_agent_warning(self):
        self.log_agent_status(
            status=V1Statuses.WARNING,
            reason="AgentLogger",
            message="The agent was interrupted, please check your deployment.",
        )

    @classmethod
    def get_healthz_config(cls) -> Optional[ChecksConfig]:
        try:
            return ChecksConfig.read(cls.HEALTH_FILE, config_type=".json")
        except Exception:  # noqa
            return None

    @classmethod
    def ping(cls):
        ChecksConfig.init_file(cls.HEALTH_FILE)
        config = cls.get_healthz_config()
        if config:
            config.last_check = now()
            config.write(cls.HEALTH_FILE, mode=config._WRITE_MODE)

    @classmethod
    def pong(cls, interval: int = 15) -> bool:
        config = cls.get_healthz_config()
        if not config:
            return False
        return not config.should_check(interval=interval)

    def refresh_executor(self):
        if (
            now() - self._executor_refreshed_at
        ).total_seconds() > settings.AGENT_CONFIG.get_executor_refresh_interval():
            logger.debug("Refreshing executor ... ")
            self.executor.refresh()
            self._executor_refreshed_at = now()

    def start(self):
        try:
            with exit_context() as exit_event:
                index = 0
                workers = get_pool_workers()

                with ThreadPoolExecutor(workers) as pool:
                    logger.debug("Thread pool Workers: {}".format(workers))
                    timeout = self.sleep_interval or get_wait(index)
                    while not exit_event.wait(timeout=timeout):
                        index += 1
                        self.refresh_executor()
                        agent_state = self.process(pool)
                        self._check_status(agent_state)
                        if agent_state.state.full:
                            index = 2
                        self.ping()
                        timeout = self.sleep_interval or get_wait(index)
                        logger.info("Sleeping for {} seconds".format(timeout))
        finally:
            self.end()

    def _check_status(self, agent_state):
        if agent_state.status == V1Statuses.STOPPED:
            print(
                "Agent has been stopped from the platform,"
                "but the deployment is still running."
                "Please either set the agent to starting or teardown the agent deployment."
            )
            self.end(sleep=self.SLEEP_STOP_TIME)
        elif agent_state.live_state < LiveState.LIVE:
            print(
                "Agent has been archived from the platform,"
                "but the deployment is still running."
                "Please either restore the agent or teardown the agent deployment."
            )
            self.end(sleep=self.SLEEP_ARCHIVED_TIME)

    def end(self, sleep: Optional[int] = None):
        self._graceful_shutdown = True
        if sleep:
            time.sleep(sleep)
        else:
            logger.info("Agent is shutting down.")

    def process(self, pool: "ThreadPoolExecutor") -> V1AgentStateResponse:
        try:
            agent_state = self.get_state()
            if agent_state.compatible_updates:
                self.sync_compatible_updates(agent_state.compatible_updates)

            if agent_state:
                logger.info("Starting runs submission process.")
            else:
                logger.info("No state was found.")
                return V1AgentStateResponse.construct()

            state = agent_state.state
            if not state:
                return agent_state
            for run_data in state.schedules or []:
                pool.submit(self.submit_run, run_data)
            for run_data in state.queued or []:
                pool.submit(self.submit_run, run_data)
            for run_data in state.checks or []:
                pool.submit(self.check_run, run_data)
            for run_data in state.stopping or []:
                pool.submit(self.stop_run, run_data)
            for run_data in state.apply or []:
                pool.submit(self.apply_run, run_data)
            for run_data in state.deleting or []:
                pool.submit(self.delete_run, run_data)
            for run_data in state.hooks or []:
                pool.submit(self.make_and_create_run, run_data)
            for run_data in state.watchdogs or []:
                pool.submit(self.make_and_create_run, run_data)
            for run_data in state.tuners or []:
                pool.submit(self.make_and_create_run, run_data, True)
            return agent_state
        except Exception as exc:
            logger.error(exc)
            return V1AgentStateResponse.construct()

    def log_run_failed(
        self,
        run_owner: str,
        run_project: str,
        run_uuid: str,
        exc: Exception,
        message: Optional[str] = None,
    ):
        message = message or "Agent failed deploying run.\n"
        message += "error: {}\n{}".format(repr(exc), traceback.format_exc())
        self.log_run_status(
            run_owner=run_owner,
            run_project=run_project,
            run_uuid=run_uuid,
            status=V1Statuses.FAILED,
            reason="AgentLogger",
            message=message,
        )
        logger.warning(message)

    def log_run_stopped(self, run_owner: str, run_project: str, run_uuid: str):
        message = "Run was not found. The agent assumed it was already stopped."
        self.log_run_status(
            run_owner=run_owner,
            run_project=run_project,
            run_uuid=run_uuid,
            status=V1Statuses.STOPPED,
            reason="AgentLogger",
            message=message,
        )
        logger.warning(message)

    def log_run_scheduled(self, run_owner: str, run_project: str, run_uuid: str):
        message = "Run was scheduled by the agent."
        self.log_run_status(
            run_owner=run_owner,
            run_project=run_project,
            run_uuid=run_uuid,
            status=V1Statuses.SCHEDULED,
            reason="AgentLogger",
            message=message,
        )
        logger.info(message)

    def log_run_running(self, run_owner: str, run_project: str, run_uuid: str):
        message = "Run changes were applied by the agent."
        self.log_run_status(
            run_owner=run_owner,
            run_project=run_project,
            run_uuid=run_uuid,
            status=V1Statuses.RUNNING,
            reason="AgentLogger",
            message=message,
        )
        logger.info(message)

    def log_run_status(
        self,
        run_owner: str,
        run_project: str,
        run_uuid: str,
        status: str,
        reason: Optional[str] = None,
        message: Optional[str] = None,
    ):
        status_condition = V1StatusCondition.get_condition(
            type=status, status=True, reason=reason, message=message
        )
        self.client.runs_v1.create_run_status(
            owner=run_owner,
            project=run_project,
            uuid=run_uuid,
            body={"condition": status_condition},
            async_req=True,
        )

    def clean_run(self, run_uuid: str, run_kind: str):
        try:
            self.executor.clean(run_uuid=run_uuid, run_kind=run_kind)
            self.executor.stop(run_uuid=run_uuid, run_kind=run_kind)
        except ApiException as e:
            if e.status == 404:
                logger.info("Run does not exist.")
        except Exception as e:
            logger.info(
                "Run could not be cleaned: {}\n{}".format(
                    repr(e), traceback.format_exc()
                )
            )

    def make_run_resource(
        self,
        owner_name: str,
        project_name: str,
        run_name: str,
        run_uuid: str,
        content: str,
        default_auth: bool = False,
    ) -> Optional[Any]:
        try:
            return self.executor.make_and_convert(
                owner_name=owner_name,
                project_name=project_name,
                run_uuid=run_uuid,
                run_name=run_name,
                content=content,
                default_auth=default_auth,
            )
        except PolyaxonConverterError as e:
            logger.info(
                "Run could not be cleaned. Agent failed converting run manifest: {}\n{}".format(
                    repr(e), traceback.format_exc()
                )
            )
        except Exception as e:
            logger.info(
                "Agent failed during compilation with unknown exception: {}\n{}".format(
                    repr(e), traceback.format_exc()
                )
            )
        return None

    def prepare_run_resource(
        self,
        owner_name: str,
        project_name: str,
        run_name: str,
        run_uuid: str,
        content: str,
    ) -> Optional[Any]:
        try:
            return self.executor.convert(
                owner_name=owner_name,
                project_name=project_name,
                run_name=run_name,
                run_uuid=run_uuid,
                content=content,
                default_auth=True,
                agent_content=self.content,
            )
        except PolyaxonConverterError as e:
            self.log_run_failed(
                run_owner=owner_name,
                run_project=project_name,
                run_uuid=run_uuid,
                exc=e,
                message="Agent failed converting run manifest.\n",
            )
        except Exception as e:
            self.log_run_failed(
                run_owner=owner_name,
                run_project=project_name,
                run_uuid=run_uuid,
                exc=e,
                message="Agent failed during compilation with unknown exception.\n",
            )
        return None

    def submit_run(self, run_data: Tuple[str, str, str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        resource = self.prepare_run_resource(
            owner_name=run_owner,
            project_name=run_project,
            run_name=run_data[2],
            run_uuid=run_uuid,
            content=run_data[3],
        )
        if not resource:
            return

        try:
            self.executor.create(
                run_uuid=run_uuid, run_kind=run_data[1], resource=resource
            )
        except ApiException as e:
            if e.status == 409:
                logger.info("Run already running, triggering an apply mechanism.")
                self.apply_run(run_data=run_data)
            else:
                logger.info("Run submission error.")
                self.log_run_failed(
                    run_owner=run_owner,
                    run_project=run_project,
                    run_uuid=run_uuid,
                    exc=e,
                )
        except Exception as e:
            self.log_run_failed(
                run_owner=run_owner,
                run_project=run_project,
                run_uuid=run_uuid,
                exc=e,
            )

    def make_and_create_run(
        self, run_data: Tuple[str, str, str, str], default_auth: bool = False
    ):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        resource = self.make_run_resource(
            owner_name=run_owner,
            project_name=run_project,
            run_name=run_data[2],
            run_uuid=run_uuid,
            content=run_data[3],
            default_auth=default_auth,
        )
        if not resource:
            return

        try:
            self.executor.create(
                run_uuid=run_uuid, run_kind=run_data[1], resource=resource
            )
        except ApiException as e:
            if e.status == 409:
                logger.info("Run already running, triggering an apply mechanism.")
            else:
                logger.info("Run submission error.")
        except Exception as e:
            logger.info(
                "Run could not be cleaned. Agent failed converting run manifest: {}\n{}".format(
                    repr(e), traceback.format_exc()
                )
            )

    def apply_run(self, run_data: Tuple[str, str, str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        resource = self.prepare_run_resource(
            owner_name=run_owner,
            project_name=run_project,
            run_name=run_data[2],
            run_uuid=run_uuid,
            content=run_data[3],
        )
        if not resource:
            return

        try:
            self.executor.apply(
                run_uuid=run_uuid, run_kind=run_data[1], resource=resource
            )
            self.log_run_running(
                run_owner=run_owner, run_project=run_project, run_uuid=run_uuid
            )
        except Exception as e:
            self.log_run_failed(
                run_owner=run_owner, run_project=run_project, run_uuid=run_uuid, exc=e
            )
            self.clean_run(run_uuid=run_uuid, run_kind=run_data[1])

    def check_run(self, run_data: Tuple[str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        try:
            self.executor.get(run_uuid=run_uuid, run_kind=run_data[1])
        except ApiException as e:
            if e.status == 404:
                logger.info(
                    "Run does not exist anymore, it could have been stopped or deleted."
                )
                self.log_run_stopped(
                    run_owner=run_owner, run_project=run_project, run_uuid=run_uuid
                )

    def stop_run(self, run_data: Tuple[str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        try:
            self.executor.stop(run_uuid=run_uuid, run_kind=run_data[1])
        except ApiException as e:
            if e.status == 404:
                logger.info("Run does not exist anymore, it could have been stopped.")
                self.log_run_stopped(
                    run_owner=run_owner, run_project=run_project, run_uuid=run_uuid
                )
        except Exception as e:
            self.log_run_failed(
                run_owner=run_owner,
                run_project=run_project,
                run_uuid=run_uuid,
                exc=e,
                message="Agent failed stopping run.\n",
            )

    def delete_run(self, run_data: Tuple[str, str, str, str]):
        run_owner, run_project, run_uuid = get_run_info(run_instance=run_data[0])
        self.clean_run(run_uuid=run_uuid, run_kind=run_data[1])
        if run_data[3]:
            self.make_and_create_run(run_data)



from typing import Dict, Optional

from polyaxon.exceptions import handle_api_error
from polyaxon.logger import logger


def handle_client_error(
    e,
    message: Optional[str] = None,
    http_messages_mapping: Optional[Dict] = None,
    sys_exit: bool = False,
):
    handle_api_error(
        e=e,
        logger=logger,
        message=message,
        http_messages_mapping=http_messages_mapping,
        sys_exit=sys_exit,
    )



META_EDGE = "edge"
META_ITERATION = "iteration"
META_BRACKET_ITERATION = "bracket_iteration"
META_REWRITE_PATH = "rewrite_path"
META_IS_EXTERNAL = "is_external"
META_HAS_EARLY_STOPPING = "has_early_stopping"
META_HAS_JOBS = "has_jobs"
META_HAS_SERVICES = "has_services"
META_HAS_DAGS = "has_dags"
META_HAS_MATRICES = "has_matrices"
META_HAS_SCHEDULES = "has_schedules"
META_HAS_HOOKS = "has_hooks"
META_IS_HOOK = "is_hook"
META_IS_ACTION = "is_action"
META_IS_TUNER = "is_tuner"
META_IS_PROMOTED = "is_promoted"
META_CONCURRENCY = "concurrency"
META_HAS_DOWNSTREAM_EVENTS_TRIGGER = "has_downstream_events_trigger"
META_UPLOAD_ARTIFACTS = "upload_artifacts"
META_COPY_ARTIFACTS = "copy_artifacts"
META_DESTINATION_IMAGE = "destination_image"
META_PROGRESS = "progress"
META_RECOMPILE = "recompile"



# To keep backwards compatibility

from typing import List, Optional

from polyaxon.client import RunClient
from traceml import tracking
from traceml.tracking import *

TRACKING_RUN = None


def init(
    owner: Optional[str] = None,
    project: Optional[str] = None,
    run_uuid: Optional[str] = None,
    client: Optional[RunClient] = None,
    track_code: bool = True,
    track_env: bool = True,
    track_logs: bool = True,
    refresh_data: bool = False,
    artifacts_path: Optional[str] = None,
    collect_artifacts: Optional[str] = None,
    collect_resources: Optional[str] = None,
    is_offline: Optional[bool] = None,
    is_new: Optional[bool] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    tags: Optional[List[str]] = None,
) -> Optional[Run]:
    global TRACKING_RUN

    TRACKING_RUN = tracking.init(
        owner=owner,
        project=project,
        run_uuid=run_uuid,
        client=client,
        track_code=track_code,
        track_env=track_env,
        track_logs=track_logs,
        refresh_data=refresh_data,
        artifacts_path=artifacts_path,
        collect_artifacts=collect_artifacts,
        collect_resources=collect_resources,
        is_offline=is_offline,
        is_new=is_new,
        name=name,
        description=description,
        tags=tags,
    )
    return TRACKING_RUN


def get_or_create_run(tracking_run: Optional[Run] = None) -> Optional[Run]:
    global TRACKING_RUN

    TRACKING_RUN = tracking.get_or_create_run(tracking_run)
    return TRACKING_RUN



from polyaxon.config.manager import ConfigManager


class ComposeConfigManager(ConfigManager):
    """Manages access cli configuration .compose file."""

    VISIBILITY = ConfigManager.Visibility.GLOBAL
    CONFIG_FILE_NAME = ".compose/.env"
    FREQUENCY = 3

    @classmethod
    def get_config_filepath(cls, create: bool = True) -> str:
        path = super().get_config_filepath(create=create)
        values = path.split("/")[:-1]
        cls._create_dir("/".join(values))
        return path



from typing import Any, Dict, Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel
from polyaxon.schemas.responses.v1_analytics_spec import V1AnalyticsSpec
from polyaxon.schemas.responses.v1_dashboard_spec import V1DashboardSpec


class V1SearchSpec(BaseResponseModel):
    query: Optional[StrictStr]
    sort: Optional[StrictStr]
    limit: Optional[int]
    offset: Optional[int]
    groupby: Optional[StrictStr]
    columns: Optional[StrictStr]
    layout: Optional[StrictStr]
    sections: Optional[StrictStr]
    compares: Optional[StrictStr]
    heat: Optional[StrictStr]
    events: Optional[V1DashboardSpec]
    histograms: Optional[Dict[str, Any]]
    trends: Optional[Dict[str, Any]]
    analytics: Optional[V1AnalyticsSpec]



from typing import Optional

from pydantic import Field, StrictInt

from polyaxon.schemas.base import BaseSchemaModel


class CeleryConfig(BaseSchemaModel):
    enabled: Optional[bool]
    task_track_started: Optional[bool] = Field(alias="taskTrackStarted")
    broker_pool_limit: Optional[StrictInt] = Field(alias="brokerPoolLimit")
    confirm_publish: Optional[bool] = Field(alias="confirmPublish")
    worker_prefetch_multiplier: Optional[StrictInt] = Field(
        alias="workerPrefetchMultiplier"
    )
    worker_max_tasks_per_child: Optional[StrictInt] = Field(
        alias="workerMaxTasksPerChild"
    )
    worker_max_memory_per_child: Optional[StrictInt] = Field(
        alias="workerMaxMemoryPerChild"
    )
    task_always_eager: Optional[bool] = Field(alias="taskAlwaysEager")



from typing import Any, Dict, Iterable, Optional

from clipped.utils.enums import get_enum_value
from clipped.utils.paths import delete_path

from polyaxon import settings
from polyaxon.auxiliaries import V1PolyaxonInitContainer, V1PolyaxonSidecarContainer
from polyaxon.compiler import resolver
from polyaxon.compiler.resolver import AgentResolver
from polyaxon.connections import V1Connection, V1ConnectionResource
from polyaxon.contexts import paths as ctx_paths
from polyaxon.exceptions import PolyaxonAgentError, PolyaxonCompilerError
from polyaxon.polyaxonfile import CompiledOperationSpecification, OperationSpecification
from polyaxon.polyflow import V1CompiledOperation
from polyaxon.runner.kinds import RunnerKind
from polyaxon.schemas.cli.agent_config import AgentConfig
from polyaxon.schemas.responses.v1_run import V1Run


class BaseExecutor:
    MIXIN_MAPPING: Dict[str, Any]
    RUNNER_KIND: RunnerKind
    CONVERTERS: Dict[str, Any]

    def __init__(self):
        self._manager = None

    @classmethod
    def _get_mixin_for_kind(cls, kind: str) -> Any:
        m = cls.MIXIN_MAPPING.get(kind)
        if not m:
            raise PolyaxonAgentError(
                "Agent received unrecognized kind {}".format(get_enum_value(kind))
            )
        return m

    @property
    def manager(self):
        if not self._manager:
            self._manager = self._get_manager()
        return self._manager

    def _get_manager(self):
        raise NotImplementedError

    def refresh(self):
        self._manager = None
        return self.manager

    def get(self, run_uuid: str, run_kind: str):
        raise NotImplementedError

    def create(self, run_uuid: str, run_kind: str, resource: Any):
        raise NotImplementedError

    def apply(self, run_uuid: str, run_kind: str, resource: Any):
        raise NotImplementedError

    def stop(self, run_uuid: str, run_kind: str):
        raise NotImplementedError

    def clean(self, run_uuid: str, run_kind: str):
        raise NotImplementedError

    def _clean_temp_execution_path(self, run_uuid: str):
        delete_path(ctx_paths.CONTEXT_TMP_RUNS_ROOT_FORMAT.format(run_uuid))

    @classmethod
    def get_resource(
        cls,
        namespace: Optional[str],
        owner_name: str,
        project_name: str,
        run_name: str,
        run_uuid: str,
        run_path: str,
        compiled_operation: V1CompiledOperation,
        artifacts_store: Optional[V1Connection],
        connection_by_names: Optional[Dict[str, V1Connection]],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        polyaxon_sidecar: Optional[V1PolyaxonSidecarContainer] = None,
        polyaxon_init: Optional[V1PolyaxonInitContainer] = None,
        default_sa: Optional[str] = None,
        internal_auth: bool = False,
        default_auth: bool = False,
    ):
        if not namespace and cls.RUNNER_KIND == RunnerKind.K8S:
            raise PolyaxonCompilerError(
                "Converter Error. "
                "Namespace is required to create a k8s resource specification."
            )
        if compiled_operation.has_pipeline:
            raise PolyaxonCompilerError(
                "Converter Error. "
                "Specification with matrix/dag/schedule section is not supported in this function."
            )

        run_kind = compiled_operation.get_run_kind()
        if run_kind not in cls.CONVERTERS:
            raise PolyaxonCompilerError(
                "Converter Error. "
                "Specification with run kind: {} is not supported in this deployment version.".format(
                    run_kind
                )
            )

        converter = cls.CONVERTERS[run_kind](
            owner_name=owner_name,
            project_name=project_name,
            run_name=run_name,
            run_uuid=run_uuid,
            namespace=namespace,
            polyaxon_init=polyaxon_init,
            polyaxon_sidecar=polyaxon_sidecar,
            internal_auth=internal_auth,
            run_path=run_path,
        )
        if converter:
            return converter.get_resource(
                compiled_operation=compiled_operation,
                artifacts_store=artifacts_store,
                connection_by_names=connection_by_names,
                secrets=secrets,
                config_maps=config_maps,
                default_sa=default_sa,
                default_auth=default_auth,
            )

    @classmethod
    def convert(
        cls,
        owner_name: str,
        project_name: str,
        run_name: str,
        run_uuid: str,
        content: str,
        default_auth: bool,
        agent_content: Optional[str] = None,
    ) -> Optional[Any]:
        agent_env = AgentResolver.construct()
        compiled_operation = CompiledOperationSpecification.read(content)

        agent_env.resolve(
            compiled_operation=compiled_operation,
            agent_config=AgentConfig.read(agent_content) if agent_content else None,
        )
        return cls.get_resource(
            owner_name=owner_name,
            project_name=project_name,
            run_name=run_name,
            run_uuid=run_uuid,
            run_path=run_uuid,
            namespace=agent_env.namespace,
            compiled_operation=compiled_operation,
            polyaxon_init=agent_env.polyaxon_init,
            polyaxon_sidecar=agent_env.polyaxon_sidecar,
            artifacts_store=agent_env.artifacts_store,
            connection_by_names=agent_env.connection_by_names,
            secrets=agent_env.secrets,
            config_maps=agent_env.config_maps,
            default_auth=default_auth,
            default_sa=agent_env.default_sa,
        )

    @classmethod
    def make_and_convert(
        cls,
        owner_name: str,
        project_name: str,
        run_uuid: str,
        run_name: str,
        content: str,
        default_sa: Optional[str] = None,
        internal_auth: bool = False,
        default_auth: bool = False,
    ) -> Optional[Any]:
        operation = OperationSpecification.read(content)
        compiled_operation = OperationSpecification.compile_operation(operation)
        resolver_obj, compiled_operation = resolver.resolve(
            compiled_operation=compiled_operation,
            owner_name=owner_name,
            project_name=project_name,
            project_uuid=project_name,
            run_name=run_name,
            run_path=run_uuid,
            run_uuid=run_uuid,
            params=operation.params,
        )
        return cls.get_resource(
            namespace=resolver_obj.namespace,
            owner_name=resolver_obj.owner_name,
            project_name=resolver_obj.project_name,
            run_name=resolver_obj.run_name,
            run_path=resolver_obj.run_path,
            run_uuid=resolver_obj.run_uuid,
            compiled_operation=compiled_operation,
            connection_by_names=resolver_obj.connection_by_names,
            internal_auth=internal_auth,
            artifacts_store=resolver_obj.artifacts_store,
            secrets=resolver_obj.secrets,
            config_maps=resolver_obj.config_maps,
            polyaxon_sidecar=resolver_obj.polyaxon_sidecar,
            polyaxon_init=resolver_obj.polyaxon_init,
            default_sa=default_sa,
            default_auth=default_auth,
        )

    def create_from_run(self, response: V1Run, default_auth: bool = False):
        resource = self.convert(
            owner_name=response.owner,
            project_name=response.project,
            run_name=response.name,
            run_uuid=response.uuid,
            default_auth=default_auth,
            content=response.content,
            agent_content=settings.AGENT_CONFIG.to_json(),
        )
        return self.create(
            run_uuid=response.uuid,
            run_kind=response.kind,
            resource=resource,
        )



from typing import List, Optional

from polyaxon.connections import V1Connection, V1ConnectionResource
from polyaxon.contexts import paths as ctx_paths
from polyaxon.docker import docker_types
from polyaxon.runner.converter import BaseConverter


class MountsMixin(BaseConverter):
    @staticmethod
    def _get_docker_volume(
        host_path: str, mount_path: str, read_only: bool
    ) -> docker_types.V1VolumeMount:
        volume = f"{host_path}:{mount_path}"
        if read_only:
            volume += ":ro"
        return docker_types.V1VolumeMount(__root__=("-v", volume))

    @staticmethod
    def _get_docker_mount(
        mount_path: str, read_only: bool
    ) -> docker_types.V1VolumeMount:
        mount = f"type=tmpfs,destination={mount_path}"
        if read_only:
            mount += ",ro"
        return docker_types.V1VolumeMount(__root__=("--mount", mount))

    @classmethod
    def _get_config_volume(cls) -> docker_types.V1VolumeMount:
        return cls._get_docker_volume(
            ctx_paths.CONTEXT_USER_POLYAXON_PATH,
            ctx_paths.CONTEXT_TMP_POLYAXON_PATH,
            True,
        )

    @classmethod
    def _get_mount_from_store(
        cls,
        store: V1Connection,
    ) -> Optional[docker_types.V1VolumeMount]:
        if not store:
            return None
        if store.is_volume_claim:
            return cls._get_docker_volume(
                store.schema_.volume_claim,
                store.schema_.mount_path,
                store.schema_.read_only,
            )

        if store.is_host_path:
            return cls._get_docker_volume(
                store.schema_.host_path,
                store.schema_.mount_path,
                store.schema_.read_only,
            )

    @classmethod
    def _get_mount_from_resource(
        cls,
        resource: V1ConnectionResource,
    ) -> Optional[docker_types.V1VolumeMount]:
        if not resource:
            return None
        if resource.mount_path:
            return cls._get_docker_volume(resource.host_path, resource.mount_path, True)

    @classmethod
    def _get_volume(
        cls,
        mount_path: str,
        host_path: Optional[str] = None,
        read_only: Optional[bool] = None,
    ) -> Optional[docker_types.V1VolumeMount]:
        if host_path:
            return cls._get_docker_volume(host_path, mount_path, read_only)

        return cls._get_docker_mount(mount_path, read_only)

    @classmethod
    def _get_docker_context_mount(cls) -> docker_types.V1VolumeMount:
        return cls._get_docker_volume(
            mount_path=ctx_paths.CONTEXT_MOUNT_DOCKER,
            host_path=ctx_paths.CONTEXT_MOUNT_DOCKER,
            read_only=True,
        )

    @classmethod
    def _get_configs_context_mount(cls) -> docker_types.V1VolumeMount:
        return cls._get_volume(
            mount_path=ctx_paths.CONTEXT_MOUNT_CONFIGS, read_only=False
        )

    @classmethod
    def _get_auth_context_mount(
        cls,
        read_only: Optional[bool] = None,
        run_path: Optional[str] = None,
    ) -> docker_types.V1VolumeMount:
        return cls._get_docker_volume(
            host_path=ctx_paths.CONTEXT_TMP_RUNS_ROOT_FORMAT.format(run_path),
            mount_path=ctx_paths.CONTEXT_MOUNT_CONFIGS,
            read_only=read_only,
        )

    @classmethod
    def _get_artifacts_context_mount(
        cls,
        read_only: bool = False,
        run_path: Optional[str] = None,
    ) -> docker_types.V1VolumeMount:
        return cls._get_docker_volume(
            host_path=ctx_paths.CONTEXT_TMP_RUNS_ROOT_FORMAT.format(run_path),
            mount_path=ctx_paths.CONTEXT_MOUNT_ARTIFACTS,
            read_only=read_only,
        )

    @classmethod
    def _get_connections_context_mount(
        cls,
        name: str,
        mount_path: str,
        run_path: str,
    ) -> docker_types.V1VolumeMount:
        return cls._get_docker_volume(
            host_path=ctx_paths.CONTEXT_TMP_RUNS_ROOT_FORMAT.format(run_path),
            mount_path=mount_path,
            read_only=False,
        )

    @classmethod
    def _get_shm_context_mount(cls) -> docker_types.V1VolumeMount:
        """
        Mount a tmpfs volume to /dev/shm.
        This will set /dev/shm size to half of the RAM of node.
        By default, /dev/shm is very small, only 64MB.
        Some experiments will fail due to lack of share memory,
        such as some experiments running on Pytorch.
        """
        return cls._get_volume(mount_path=ctx_paths.CONTEXT_MOUNT_SHM, read_only=False)

    @classmethod
    def _get_mounts(
        cls,
        use_auth_context: bool,
        use_docker_context: bool,
        use_shm_context: bool,
        use_artifacts_context: bool,
        run_path: Optional[str] = None,
    ) -> List[docker_types.V1VolumeMount]:
        mounts = []
        if use_auth_context:
            mounts.append(
                cls._get_auth_context_mount(read_only=True, run_path=run_path)
            )
        if use_artifacts_context:
            mounts.append(
                cls._get_artifacts_context_mount(read_only=False, run_path=run_path)
            )
        if use_docker_context:
            mounts.append(cls._get_docker_context_mount())
        if use_shm_context:
            mounts.append(cls._get_shm_context_mount())

        return mounts



from typing import Optional

from pydantic import Field, StrictInt, StrictStr

from polyaxon.schemas.base import BaseSchemaModel


class ProxyConfig(BaseSchemaModel):
    enabled: Optional[bool]
    use_in_ops: Optional[bool] = Field(alias="useInOps")
    http_proxy: Optional[StrictStr] = Field(alias="httpProxy")
    https_proxy: Optional[StrictStr] = Field(alias="httpsProxy")
    no_proxy: Optional[StrictStr] = Field(alias="noProxy")
    port: Optional[StrictInt]
    host: Optional[StrictStr]
    kind: Optional[StrictStr]



from typing import Dict, List, Optional, Union

from pydantic import Field, StrictStr, validator

from polyaxon.containers.pull_policy import PullPolicy
from polyaxon.k8s import k8s_schemas, k8s_validation
from polyaxon.schemas.base import BaseSchemaModel


class BaseServiceConfig(BaseSchemaModel):
    _SWAGGER_FIELDS = {
        "resources",
        "affinity",
        "tolerations",
    }

    image: Optional[StrictStr]
    image_tag: Optional[StrictStr] = Field(alias="imageTag")
    image_pull_policy: Optional[PullPolicy] = Field(alias="imagePullPolicy")
    resources: Optional[Union[k8s_schemas.V1ResourceRequirements, Dict]]
    node_selector: Optional[Dict[StrictStr, StrictStr]] = Field(alias="nodeSelector")
    affinity: Optional[Union[k8s_schemas.V1Affinity, Dict]]
    tolerations: Optional[List[Union[k8s_schemas.V1Toleration, Dict]]]
    image_pull_secrets: Optional[List[StrictStr]] = Field(alias="imagePullSecrets")
    hpa: Optional[Dict]

    @validator("resources", always=True, pre=True)
    def validate_resources(cls, v):
        return k8s_validation.validate_k8s_resource_requirements(v)

    @validator("affinity", always=True, pre=True)
    def validate_affinity(cls, v):
        return k8s_validation.validate_k8s_affinity(v)

    @validator("tolerations", always=True, pre=True)
    def validate_tolerations(cls, v):
        if not v:
            return v
        return [k8s_validation.validate_k8s_toleration(vi) for vi in v]



from typing import Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseSchemaModel


class V1Version(BaseSchemaModel):
    min: Optional[StrictStr]
    latest: Optional[StrictStr]



from typing import List, Optional, Tuple, Union

from clipped.utils.enums import get_enum_value
from clipped.utils.lists import to_list

from polyaxon.auxiliaries import V1PolyaxonInitContainer
from polyaxon.connections import V1Connection, V1ConnectionKind
from polyaxon.constants.globals import DEFAULT
from polyaxon.containers.names import (
    INIT_ARTIFACTS_CONTAINER_PREFIX,
    INIT_AUTH_CONTAINER,
    INIT_CUSTOM_CONTAINER_PREFIX,
    INIT_DOCKERFILE_CONTAINER_PREFIX,
    INIT_FILE_CONTAINER_PREFIX,
    INIT_GIT_CONTAINER_PREFIX,
    INIT_TENSORBOARD_CONTAINER_PREFIX,
    generate_container_name,
)
from polyaxon.contexts import paths as ctx_paths
from polyaxon.env_vars.keys import EV_KEYS_SSH_PATH
from polyaxon.exceptions import PolyaxonConverterError
from polyaxon.k8s import k8s_schemas
from polyaxon.polyflow import V1Plugins
from polyaxon.runner.converter import BaseConverter as _BaseConverter
from polyaxon.runner.converter.common import constants
from polyaxon.runner.converter.common.volumes import get_volume_name
from polyaxon.runner.converter.init.artifacts import init_artifact_context_args
from polyaxon.runner.converter.init.file import FILE_INIT_COMMAND, get_file_init_args
from polyaxon.runner.converter.init.git import REPO_INIT_COMMAND, get_repo_context_args
from polyaxon.runner.converter.init.store import get_volume_args
from polyaxon.runner.converter.init.tensorboard import (
    TENSORBOARD_INIT_COMMAND,
    get_tensorboard_args,
)
from polyaxon.schemas.types import (
    V1ArtifactsType,
    V1DockerfileType,
    V1FileType,
    V1TensorboardType,
)


class InitConverter(_BaseConverter):
    @classmethod
    def _get_base_store_container(
        cls,
        container: Optional[k8s_schemas.V1Container],
        container_name: str,
        polyaxon_init: V1PolyaxonInitContainer,
        store: V1Connection,
        env: List[k8s_schemas.V1EnvVar],
        env_from: List[k8s_schemas.V1EnvFromSource],
        volume_mounts: List[k8s_schemas.V1VolumeMount],
        args: List[str],
        command: Optional[List[str]] = None,
    ) -> Optional[k8s_schemas.V1Container]:
        env = env or []
        env_from = env_from or []
        volume_mounts = volume_mounts or []

        # Artifact store needs to allow init the contexts as well, so the store is not required
        if not store:
            raise PolyaxonConverterError("Init store container requires a store")

        if store.is_bucket:
            secret = store.secret
            volume_mounts = volume_mounts + to_list(
                cls._get_mount_from_resource(resource=secret), check_none=True
            )
            env = env + to_list(
                cls._get_items_from_secret(secret=secret), check_none=True
            )
            env_from = env_from + to_list(
                cls._get_env_from_secret(secret=secret), check_none=True
            )
            config_map = store.config_map
            volume_mounts = volume_mounts + to_list(
                cls._get_mount_from_resource(resource=config_map), check_none=True
            )
            env = env + to_list(
                cls._get_items_from_config_map(config_map=config_map), check_none=True
            )
            env_from = env_from + to_list(
                cls._get_env_from_config_map(config_map=config_map), check_none=True
            )
        else:
            volume_mounts = volume_mounts + to_list(
                cls._get_mount_from_store(store=store), check_none=True
            )
        # Add connections catalog env vars information
        env += to_list(
            cls._get_connections_catalog_env_var(connections=[store]),
            check_none=True,
        )
        env += to_list(cls._get_connection_env_var(connection=store), check_none=True)

        return cls._patch_container(
            container=container,
            name=container_name,
            image=polyaxon_init.get_image(),
            image_pull_policy=polyaxon_init.image_pull_policy,
            command=command or ["/bin/sh", "-c"],
            args=args,
            env=env,
            env_from=env_from,
            resources=polyaxon_init.get_resources(),
            volume_mounts=volume_mounts,
        )

    @classmethod
    def _get_custom_init_container(
        cls,
        connection: V1Connection,
        plugins: V1Plugins,
        container: Optional[k8s_schemas.V1Container],
        run_path: str,
        env: List[k8s_schemas.V1EnvVar] = None,
        mount_path: Optional[str] = None,
    ) -> k8s_schemas.V1Container:
        if not connection:
            raise PolyaxonConverterError(
                "A connection is required to create a repo context."
            )

        volume_name = (
            get_volume_name(mount_path)
            if mount_path
            else constants.VOLUME_MOUNT_ARTIFACTS
        )
        mount_path = mount_path or ctx_paths.CONTEXT_MOUNT_ARTIFACTS
        volume_mounts = [
            cls._get_connections_context_mount(
                name=volume_name, mount_path=mount_path, run_path=run_path
            )
        ]

        if plugins and plugins.auth:
            volume_mounts.append(
                cls._get_auth_context_mount(run_path=run_path, read_only=True)
            )

        env = to_list(env, check_none=True)
        env_from = []
        secret = connection.secret
        if secret:
            volume_mounts += to_list(
                cls._get_mount_from_resource(resource=secret), check_none=True
            )
            env += to_list(cls._get_items_from_secret(secret=secret), check_none=True)
            env_from += to_list(
                cls._get_env_from_secret(secret=secret), check_none=True
            )

        # Add connections catalog env vars information
        env += to_list(
            cls._get_connections_catalog_env_var(connections=[connection]),
            check_none=True,
        )
        env += to_list(
            cls._get_connection_env_var(connection=connection), check_none=True
        )
        config_map = connection.config_map
        if config_map:
            volume_mounts += to_list(
                cls._get_mount_from_resource(resource=config_map), check_none=True
            )
            env += to_list(
                cls._get_items_from_config_map(config_map=config_map), check_none=True
            )
            env_from = to_list(
                cls._get_env_from_config_map(config_map=config_map), check_none=True
            )
        container_name = container.name or generate_container_name(
            INIT_CUSTOM_CONTAINER_PREFIX, connection.name
        )
        return cls._patch_container(
            container=container,
            name=container_name,
            env=env,
            env_from=env_from,
            volume_mounts=volume_mounts,
        )

    @classmethod
    def _get_dockerfile_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        dockerfile_args: V1DockerfileType,
        plugins: V1Plugins,
        run_path: str,
        run_instance: str,
        container: Optional[k8s_schemas.V1Container] = None,
        env: List[k8s_schemas.V1EnvVar] = None,
        mount_path: Optional[str] = None,
    ) -> k8s_schemas.V1Container:
        env = to_list(env, check_none=True)
        env = env + [cls._get_run_instance_env_var(run_instance)]

        container_name = generate_container_name(INIT_DOCKERFILE_CONTAINER_PREFIX)
        if not container:
            container = cls._new_container(name=container_name)

        volume_name = (
            get_volume_name(mount_path)
            if mount_path
            else constants.VOLUME_MOUNT_ARTIFACTS
        )
        mount_path = mount_path or ctx_paths.CONTEXT_MOUNT_ARTIFACTS
        volume_mounts = [
            cls._get_connections_context_mount(
                name=volume_name, mount_path=mount_path, run_path=run_path
            )
        ]
        if plugins and plugins.auth:
            volume_mounts.append(
                cls._get_auth_context_mount(run_path=run_path, read_only=True)
            )

        return cls._patch_container(
            container=container,
            name=container_name,
            image=polyaxon_init.get_image(),
            image_pull_policy=polyaxon_init.image_pull_policy,
            command=["polyaxon", "docker", "generate"],
            args=[
                "--build-context={}".format(dockerfile_args.to_json()),
                "--destination={}".format(mount_path),
                "--copy-path={}".format(
                    ctx_paths.CONTEXT_MOUNT_RUN_OUTPUTS_FORMAT.format(run_path)
                ),
                "--track",
            ],
            env=env,
            volume_mounts=volume_mounts,
            resources=polyaxon_init.get_resources(),
        )

    @classmethod
    def _get_file_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        file_args: V1FileType,
        plugins: V1Plugins,
        run_path: str,
        run_instance: str,
        container: Optional[k8s_schemas.V1Container] = None,
        env: List[k8s_schemas.V1EnvVar] = None,
        mount_path: Optional[str] = None,
    ) -> k8s_schemas.V1Container:
        env = to_list(env, check_none=True)
        env = env + [cls._get_run_instance_env_var(run_instance)]

        container_name = generate_container_name(INIT_FILE_CONTAINER_PREFIX)
        if not container:
            container = cls._new_container(name=container_name)

        volume_name = (
            get_volume_name(mount_path)
            if mount_path
            else constants.VOLUME_MOUNT_ARTIFACTS
        )
        mount_path = mount_path or ctx_paths.CONTEXT_MOUNT_ARTIFACTS
        volume_mounts = [
            cls._get_connections_context_mount(
                name=volume_name, mount_path=mount_path, run_path=run_path
            )
        ]
        if plugins and plugins.auth:
            volume_mounts.append(
                cls._get_auth_context_mount(run_path=run_path, read_only=True)
            )

        file_args.filename = file_args.filename or "file"
        return cls._patch_container(
            container=container,
            name=container_name,
            image=polyaxon_init.get_image(),
            image_pull_policy=polyaxon_init.image_pull_policy,
            command=FILE_INIT_COMMAND,
            args=get_file_init_args(
                file_args=file_args, run_path=run_path, mount_path=mount_path
            ),
            env=env,
            volume_mounts=volume_mounts,
            resources=polyaxon_init.get_resources(),
        )

    @classmethod
    def _get_git_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        connection: V1Connection,
        plugins: V1Plugins,
        run_path: str,
        container: Optional[k8s_schemas.V1Container] = None,
        env: List[k8s_schemas.V1EnvVar] = None,
        mount_path: Optional[str] = None,
        track: bool = False,
    ) -> k8s_schemas.V1Container:
        if not connection:
            raise PolyaxonConverterError(
                "A connection is required to create a repo context."
            )
        container_name = generate_container_name(
            INIT_GIT_CONTAINER_PREFIX, connection.name
        )
        if not container:
            container = cls._new_container(name=container_name)

        volume_name = (
            get_volume_name(mount_path)
            if mount_path
            else constants.VOLUME_MOUNT_ARTIFACTS
        )
        mount_path = mount_path or ctx_paths.CONTEXT_MOUNT_ARTIFACTS
        volume_mounts = [
            cls._get_connections_context_mount(
                name=volume_name, mount_path=mount_path, run_path=run_path
            )
        ]

        if plugins and plugins.auth:
            volume_mounts.append(
                cls._get_auth_context_mount(run_path=run_path, read_only=True)
            )

        env = to_list(env, check_none=True)
        env_from = []
        secret = connection.secret
        if secret:
            volume_mounts += to_list(
                cls._get_mount_from_resource(resource=secret), check_none=True
            )
            env += to_list(cls._get_items_from_secret(secret=secret), check_none=True)
            env_from += to_list(
                cls._get_env_from_secret(secret=secret), check_none=True
            )

        # Add connections catalog env vars information
        env += to_list(
            cls._get_connections_catalog_env_var(connections=[connection]),
            check_none=True,
        )
        env += to_list(
            cls._get_connection_env_var(connection=connection), check_none=True
        )
        # Add special handling to auto-inject ssh mount path
        if connection.kind == V1ConnectionKind.SSH and secret.mount_path:
            env += [cls._get_env_var(EV_KEYS_SSH_PATH, secret.mount_path)]
        config_map = connection.config_map
        if config_map:
            volume_mounts += to_list(
                cls._get_mount_from_resource(resource=config_map), check_none=True
            )
            env += to_list(
                cls._get_items_from_config_map(config_map=config_map), check_none=True
            )
            env_from = to_list(
                cls._get_env_from_config_map(config_map=config_map), check_none=True
            )
        args = get_repo_context_args(
            name=connection.name,
            # Handle the case of custom connection
            url=getattr(connection.schema_, "url", None),
            revision=getattr(connection.schema_, "revision", None),
            flags=getattr(connection.schema_, "flags", None),
            mount_path=mount_path,
            connection=connection.name if track else None,
        )
        return cls._patch_container(
            container=container,
            name=container_name,
            image=polyaxon_init.get_image(),
            image_pull_policy=polyaxon_init.image_pull_policy,
            command=REPO_INIT_COMMAND,
            args=args,
            env=env,
            env_from=env_from,
            volume_mounts=volume_mounts,
            resources=polyaxon_init.get_resources(),
        )

    @classmethod
    def _get_store_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        connection: V1Connection,
        artifacts: V1ArtifactsType,
        paths: Union[List[str], List[Tuple[str, str]]],
        run_path: str,
        container: Optional[k8s_schemas.V1Container] = None,
        env: List[k8s_schemas.V1EnvVar] = None,
        mount_path: Optional[str] = None,
        is_default_artifacts_store: bool = False,
    ) -> k8s_schemas.V1Container:
        container_name = generate_container_name(
            INIT_ARTIFACTS_CONTAINER_PREFIX, connection.name
        )
        if not container:
            container = cls._new_container(name=container_name)

        volume_name = (
            get_volume_name(mount_path)
            if mount_path
            else constants.VOLUME_MOUNT_ARTIFACTS
        )
        volume_mount_path = mount_path or ctx_paths.CONTEXT_MOUNT_ARTIFACTS
        volume_mounts = [
            cls._get_connections_context_mount(
                name=volume_name, mount_path=volume_mount_path, run_path=run_path
            )
        ]
        mount_path = mount_path or (
            ctx_paths.CONTEXT_MOUNT_ARTIFACTS
            if is_default_artifacts_store
            else ctx_paths.CONTEXT_MOUNT_ARTIFACTS_FORMAT.format(connection.name)
        )

        return cls._get_base_store_container(
            container=container,
            container_name=container_name,
            polyaxon_init=polyaxon_init,
            store=connection,
            env=env,
            env_from=[],
            volume_mounts=volume_mounts,
            args=[
                get_volume_args(
                    store=connection,
                    mount_path=mount_path,
                    artifacts=artifacts,
                    paths=paths,
                )
            ],
        )

    @classmethod
    def _get_tensorboard_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        artifacts_store: V1Connection,
        tb_args: V1TensorboardType,
        plugins: V1Plugins,
        run_path: str,
        run_instance: str,
        container: Optional[k8s_schemas.V1Container] = None,
        env: List[k8s_schemas.V1EnvVar] = None,
        mount_path: Optional[str] = None,
    ) -> k8s_schemas.V1Container:
        env = to_list(env, check_none=True)
        env = env + [cls._get_run_instance_env_var(run_instance)]

        container_name = generate_container_name(INIT_TENSORBOARD_CONTAINER_PREFIX)
        if not container:
            container = cls._new_container(name=container_name)

        volume_name = (
            get_volume_name(mount_path)
            if mount_path
            else constants.VOLUME_MOUNT_ARTIFACTS
        )
        mount_path = mount_path or ctx_paths.CONTEXT_MOUNT_ARTIFACTS
        volume_mounts = [
            cls._get_connections_context_mount(
                name=volume_name, mount_path=mount_path, run_path=run_path
            )
        ]
        if plugins and plugins.auth:
            volume_mounts.append(
                cls._get_auth_context_mount(run_path=run_path, read_only=True)
            )

        args = get_tensorboard_args(
            tb_args=tb_args,
            context_from=artifacts_store.store_path,
            context_to=mount_path,
            connection_kind=get_enum_value(artifacts_store.kind),
        )

        return cls._get_base_store_container(
            container=container,
            container_name=container_name,
            polyaxon_init=polyaxon_init,
            store=artifacts_store,
            command=TENSORBOARD_INIT_COMMAND,
            args=args,
            env=env,
            env_from=[],
            volume_mounts=volume_mounts,
        )

    @classmethod
    def _get_auth_context_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        run_path: str,
        env: Optional[List[k8s_schemas.V1EnvVar]] = None,
    ) -> k8s_schemas.V1Container:
        env = to_list(env, check_none=True)
        container = k8s_schemas.V1Container(
            name=INIT_AUTH_CONTAINER,
            image=polyaxon_init.get_image(),
            image_pull_policy=polyaxon_init.image_pull_policy,
            command=["polyaxon", "initializer", "auth"],
            env=env,
            resources=polyaxon_init.get_resources(),
            volume_mounts=[
                cls._get_auth_context_mount(run_path=run_path, read_only=False)
            ],
        )
        return cls._patch_container(container)

    @classmethod
    def _get_artifacts_path_init_container(
        cls,
        polyaxon_init: V1PolyaxonInitContainer,
        artifacts_store: V1Connection,
        run_path: str,
        auto_resume: bool,
        env: Optional[List[k8s_schemas.V1EnvVar]] = None,
    ) -> k8s_schemas.V1Container:
        if not artifacts_store:
            raise PolyaxonConverterError("Init artifacts container requires a store.")

        env = to_list(env, check_none=True)
        init_args = init_artifact_context_args(run_path=run_path)
        if auto_resume:
            init_args.append(
                get_volume_args(
                    store=artifacts_store,
                    mount_path=ctx_paths.CONTEXT_MOUNT_ARTIFACTS,
                    artifacts=V1ArtifactsType(dirs=[run_path]),
                    paths=None,
                    sync_fw=True,
                )
            )

        container_name = generate_container_name(
            INIT_ARTIFACTS_CONTAINER_PREFIX, DEFAULT, False
        )
        container = cls._new_container(name=container_name)

        return cls._get_base_store_container(
            container_name=container_name,
            container=container,
            polyaxon_init=polyaxon_init,
            store=artifacts_store,
            env=env,
            env_from=[],
            volume_mounts=[cls._get_artifacts_context_mount(run_path=run_path)],
            # If we are dealing with a volume we need to make sure the path exists for the user
            # We also clean the path if this is not a resume run
            args=[" ".join(init_args)],
        )



from clipped.types.gcs import GcsPath
from clipped.types.s3 import S3Path
from clipped.types.uri import Uri
from clipped.types.wasb import WasbPath

from polyaxon.schemas.types.artifacts import V1ArtifactsType
from polyaxon.schemas.types.dockerfile import V1DockerfileType
from polyaxon.schemas.types.event import V1EventType
from polyaxon.schemas.types.file import V1FileType
from polyaxon.schemas.types.git import V1GitType
from polyaxon.schemas.types.tensorboard import V1TensorboardType

V1GcsType = GcsPath
V1S3Type = S3Path
V1UriType = Uri
V1WasbType = WasbPath



from typing_extensions import Literal

from pydantic import StrictStr

from polyaxon.polyflow.references.mixin import RefMixin
from polyaxon.schemas.base import BaseSchemaModel


class V1UrlRef(BaseSchemaModel, RefMixin):
    _IDENTIFIER = "url_ref"
    _USE_DISCRIMINATOR = True

    kind: Literal[_IDENTIFIER] = _IDENTIFIER
    url: StrictStr

    def get_kind_value(self):
        return self.url



from typing import Dict

from pydantic import ValidationError

from polyaxon.exceptions import PolyaxonValidationError
from polyaxon.polyflow.run.cleaner import V1CleanerJob
from polyaxon.polyflow.run.dag import V1Dag
from polyaxon.polyflow.run.dask import V1DaskJob, V1DaskReplica
from polyaxon.polyflow.run.job import V1Job
from polyaxon.polyflow.run.kinds import V1RunKind
from polyaxon.polyflow.run.kubeflow.mpi_job import V1MPIJob
from polyaxon.polyflow.run.kubeflow.mx_job import V1MXJob
from polyaxon.polyflow.run.kubeflow.paddle_job import V1PaddleJob
from polyaxon.polyflow.run.kubeflow.pytorch_job import V1PytorchJob
from polyaxon.polyflow.run.kubeflow.replica import V1KFReplica
from polyaxon.polyflow.run.kubeflow.tf_job import V1TFJob
from polyaxon.polyflow.run.kubeflow.xgboost_job import V1XGBoostJob
from polyaxon.polyflow.run.notifier import V1NotifierJob
from polyaxon.polyflow.run.ray import V1RayJob, V1RayReplica
from polyaxon.polyflow.run.service import V1Service
from polyaxon.polyflow.run.spark import V1SparkJob, V1SparkReplica
from polyaxon.polyflow.run.tuner import V1TunerJob


def validate_run_patch(run_patch: Dict, kind: V1RunKind):
    if kind == V1RunKind.JOB:
        patch = V1Job.from_dict(run_patch)
    elif kind == V1RunKind.SERVICE:
        patch = V1Service.from_dict(run_patch)
    elif kind == V1RunKind.DAG:
        patch = V1Dag.from_dict(run_patch)
    elif kind == V1RunKind.MPIJOB:
        try:
            patch = V1MPIJob.from_dict(run_patch)
        except ValidationError:
            patch = V1KFReplica.from_dict(run_patch)
    elif kind == V1RunKind.PYTORCHJOB:
        try:
            patch = V1PytorchJob.from_dict(run_patch)
        except ValidationError:
            patch = V1KFReplica.from_dict(run_patch)
    elif kind == V1RunKind.PADDLEJOB:
        try:
            patch = V1PaddleJob.from_dict(run_patch)
        except ValidationError:
            patch = V1KFReplica.from_dict(run_patch)
    elif kind == V1RunKind.TFJOB:
        try:
            patch = V1TFJob.from_dict(run_patch)
        except ValidationError:
            patch = V1KFReplica.from_dict(run_patch)
    elif kind == V1RunKind.MXJOB:
        try:
            patch = V1MXJob.from_dict(run_patch)
        except ValidationError:
            patch = V1KFReplica.from_dict(run_patch)
    elif kind == V1RunKind.XGBJOB:
        try:
            patch = V1XGBoostJob.from_dict(run_patch)
        except ValidationError:
            patch = V1KFReplica.from_dict(run_patch)
    elif kind == V1RunKind.SPARKJOB:
        try:
            patch = V1SparkJob.from_dict(run_patch)
        except ValidationError:
            patch = V1SparkReplica.from_dict(run_patch)
    elif kind == V1RunKind.RAYJOB:
        try:
            patch = V1RayJob.from_dict(run_patch)
        except ValidationError:
            patch = V1RayReplica.from_dict(run_patch)
    elif kind == V1RunKind.DASKJOB:
        try:
            patch = V1DaskJob.from_dict(run_patch)
        except ValidationError:
            patch = V1DaskReplica.from_dict(run_patch)
    elif kind == V1RunKind.NOTIFIER:
        patch = V1NotifierJob.from_dict(run_patch)
    elif kind == V1RunKind.TUNER:
        patch = V1TunerJob.from_dict(run_patch)
    elif kind == V1RunKind.CLEANER:
        patch = V1CleanerJob.from_dict(run_patch)
    else:
        raise PolyaxonValidationError(
            "runPatch cannot be validate without a supported kind."
        )

    return patch



from polyaxon.config.spec import ConfigSpec
from polyaxon.deploy.schemas.deployment import DeploymentConfig


def read(filepaths):
    data = ConfigSpec.read_from(filepaths)
    return DeploymentConfig.from_dict(data)



from typing import Dict, List, Optional

from kubernetes import client

from polyaxon.k8s import k8s_schemas
from polyaxon.k8s.converter.pod.spec import get_pod_spec, get_pod_template_spec
from polyaxon.k8s.custom_resources.operation import get_operation_custom_object
from polyaxon.k8s.custom_resources.setter import (
    set_collect_logs,
    set_notify,
    set_sync_statuses,
    set_termination,
)
from polyaxon.k8s.replica import ReplicaSpec
from polyaxon.polyflow import V1Notification, V1Termination


def get_dask_replicas_template(
    namespace: str,
    resource_name: str,
    replica_name: str,
    replica: Optional[ReplicaSpec],
    labels: Dict[str, str],
    annotations: Dict[str, str],
    template_spec: Dict,
    args: Optional[List[str]] = None,
    ports: Optional[List[k8s_schemas.V1ContainerPort]] = None,
    readiness_probe: Optional[client.V1Probe] = None,
    liveness_probe: Optional[client.V1Probe] = None,
):
    if not replica:
        return

    if ports and not replica.main_container.ports:
        replica.main_container.ports = ports
    if args and replica.main_container.args is None:
        replica.main_container.args = args
    if readiness_probe and replica.main_container.readiness_probe is None:
        replica.main_container.readiness_probe = readiness_probe
    if liveness_probe and replica.main_container.liveness_probe is None:
        replica.main_container.liveness_probe = liveness_probe

    metadata, pod_spec = get_pod_spec(
        namespace=namespace,
        main_container=replica.main_container,
        sidecar_containers=replica.sidecar_containers,
        init_containers=replica.init_containers,
        resource_name=resource_name,
        volumes=replica.volumes,
        environment=replica.environment,
        labels=labels,
        annotations=annotations,
    )

    template_spec[replica_name] = {
        "replicas": replica.num_replicas,
        "restartPolicy": pod_spec.restart_policy or "Never",
        "template": get_pod_template_spec(metadata=metadata, pod_spec=pod_spec),
    }


def get_dask_job_custom_resource(
    resource_name: str,
    namespace: str,
    job: Optional[ReplicaSpec],
    worker: Optional[ReplicaSpec],
    scheduler: Optional[ReplicaSpec],
    termination: V1Termination,
    collect_logs: bool,
    sync_statuses: bool,
    notifications: List[V1Notification],
    labels: Dict[str, str],
    annotations: Dict[str, str],
) -> Dict:
    template_spec = {}

    get_dask_replicas_template(
        replica_name="Job",
        replica=job,
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        template_spec=template_spec,
    )
    get_dask_replicas_template(
        replica_name="Worker",
        replica=worker,
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        template_spec=template_spec,
        args=[
            "dask-worker",
            "--name",
            "$(DASK_WORKER_NAME)",
            "--dashboard",
            "--dashboard-address",
            "8788",
        ],
        ports=[
            k8s_schemas.V1ContainerPort(
                name="http-dashboard", container_port=8788, protocol="TCP"
            )
        ],
    )
    get_dask_replicas_template(
        replica_name="Scheduler",
        replica=scheduler,
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        template_spec=template_spec,
        args=["dask-scheduler"],
        ports=[
            k8s_schemas.V1ContainerPort(
                name="tcp-comm", container_port=8786, protocol="TCP"
            ),
            k8s_schemas.V1ContainerPort(
                name="http-dashboard", container_port=8787, protocol="TCP"
            ),
        ],
        readiness_probe=client.V1Probe(
            http_get=client.V1HTTPGetAction(path="/health", port="http-dashboard"),
            initial_delay_seconds=5,
            period_seconds=5,
        ),
        liveness_probe=client.V1Probe(
            http_get=client.V1HTTPGetAction(path="/health", port="http-dashboard"),
            initial_delay_seconds=15,
            period_seconds=15,
        ),
    )
    service = client.V1ServiceSpec(
        type="ClusterIP",
        selector={
            **labels,
            "dask.org/cluster-name": resource_name,
            "dask.org/component": "scheduler",
        },
        ports=[
            client.V1ServicePort(
                port=8786, target_port="tcp-comm", name="tcp-comm", protocol="TCP"
            ),
            client.V1ServicePort(
                port=8787,
                target_port="http-dashboard",
                name="http-dashboard",
                protocol="TCP",
            ),
        ],
    )
    template_spec = {"replicaSpecs": template_spec, "service": service}
    custom_object = {"daskJobSpec": template_spec}
    custom_object = set_termination(
        custom_object=custom_object, termination=termination
    )
    custom_object = set_collect_logs(
        custom_object=custom_object, collect_logs=collect_logs
    )
    custom_object = set_sync_statuses(
        custom_object=custom_object, sync_statuses=sync_statuses
    )
    custom_object = set_notify(custom_object=custom_object, notifications=notifications)

    return get_operation_custom_object(
        namespace=namespace,
        resource_name=resource_name,
        labels=labels,
        annotations=annotations,
        custom_object=custom_object,
    )



from typing import Dict

from polyaxon.k8s.executor.base import BaseExecutor
from polyaxon.k8s.manager.async_manager import AsyncK8sManager
from polyaxon.utils.fqn_utils import get_resource_name


class AsyncExecutor(BaseExecutor):
    def _get_manager(self):
        return AsyncK8sManager(
            namespace=self.namespace,
            in_cluster=self.in_cluster,
        )

    async def create(self, run_uuid: str, run_kind: str, resource: Dict):
        mixin = self._get_mixin_for_kind(kind=run_kind)
        resource_name = get_resource_name(run_uuid)
        return await self.manager.create_custom_object(
            name=resource_name,
            group=mixin.GROUP,
            version=mixin.API_VERSION,
            plural=mixin.PLURAL,
            body=resource,
        )

    async def apply(self, run_uuid: str, run_kind: str, resource: Dict):
        mixin = self._get_mixin_for_kind(kind=run_kind)
        resource_name = get_resource_name(run_uuid)
        return await self.manager.update_custom_object(
            name=resource_name,
            group=mixin.GROUP,
            version=mixin.API_VERSION,
            plural=mixin.PLURAL,
            body=resource,
        )

    async def stop(self, run_uuid: str, run_kind: str):
        mixin = self._get_mixin_for_kind(kind=run_kind)
        resource_name = get_resource_name(run_uuid)
        return await self.manager.delete_custom_object(
            name=resource_name,
            group=mixin.GROUP,
            version=mixin.API_VERSION,
            plural=mixin.PLURAL,
        )

    async def get(self, run_uuid: str, run_kind: str):
        mixin = self._get_mixin_for_kind(kind=run_kind)
        resource_name = get_resource_name(run_uuid)
        return await self.manager.get_custom_object(
            name=resource_name,
            group=mixin.GROUP,
            version=mixin.API_VERSION,
            plural=mixin.PLURAL,
        )



from polyaxon.k8s.events import get_container_status, get_container_statuses_by_name
from polyaxon.k8s.pods import PodLifeCycle


def is_container_terminated(status, container_id):
    container_statuses = status.get("container_statuses") or []
    statuses_by_name = get_container_statuses_by_name(container_statuses)
    statuses = get_container_status(statuses_by_name, (container_id,))
    statuses = statuses or {}
    return statuses.get("state", {}).get("terminated")


def is_pod_running(event, container_id):
    event = event.to_dict()
    event_status = event.get("status", {})
    is_terminated = is_container_terminated(
        status=event_status, container_id=container_id
    )

    return (
        event_status.get("phase")
        in {PodLifeCycle.RUNNING, PodLifeCycle.PENDING, PodLifeCycle.CONTAINER_CREATING}
        and not is_terminated
    )



import os

from queue import Queue
from time import sleep, time

from polyaxon import settings
from polyaxon.client.workers.base_worker import BaseWorker
from polyaxon.logger import logger


class QueueWorker(BaseWorker):
    TIMEOUT_ATTEMPTS = 5
    QUEUE_SIZE = -1  # inf
    END_EVENT = object()
    NAME = "polyaxon.QueueWorker"

    def __init__(self, timeout=None, queue_size=None):
        super().__init__()
        self._queue = Queue(queue_size or self.QUEUE_SIZE)
        self._timeout = (
            timeout if timeout is not None else settings.CLIENT_CONFIG.timeout
        )

    def atexit(self):
        with self._lock:
            if not self.is_alive():
                return

            self._queue.put_nowait(self.END_EVENT)

            def timeout_join(timeout, queue):
                end = time() + timeout
                queue.all_tasks_done.acquire()
                try:
                    while queue.unfinished_tasks:
                        current_timeout = end - time()
                        if current_timeout <= 0:
                            # timed out
                            return False

                        queue.all_tasks_done.wait(timeout=current_timeout)

                    return True

                finally:
                    queue.all_tasks_done.release()

            # ensure wait
            timeout = min(settings.MIN_TIMEOUT, self._timeout / self.TIMEOUT_ATTEMPTS)
            if timeout_join(timeout=timeout, queue=self._queue):
                timeout = 0
            else:
                # Queue still has message, try another time
                size = self._queue.qsize()

                if not settings.CLIENT_CONFIG.is_managed:
                    print(
                        "Polyaxon %s is attempting to send %i pending messages"
                        % (self.NAME, size)
                    )
                    print("Waiting up to {} seconds".format(self._timeout))
                    if os.name == "nt":
                        print("Press Ctrl-Break to quit")
                    else:
                        print("Press Ctrl-C to quit")

            sleep(settings.MIN_TIMEOUT)  # Allow tasks to get executed
            while timeout > 0 and not timeout_join(timeout=timeout, queue=self._queue):
                timeout = min(
                    timeout + self._timeout / self.TIMEOUT_ATTEMPTS,
                    self._timeout - timeout,
                )

            size = self._queue.qsize()
            if size > 0:
                print(
                    "Polyaxon %s timed out and did not manage to send %i messages"
                    % (self.NAME, size)
                )

            self._thread = None

    def stop(self, timeout=None):
        with self._lock:
            if self._thread:
                self._queue.put_nowait(self.END_EVENT)
                self._thread.join(timeout=timeout)
                self._thread = None
                self._thread_for_pid = None

    def queue(self, callback, *args, **kwargs):
        self.is_running()
        self._queue.put_nowait((callback, args, kwargs))

    def _target(self):
        while True:
            record = self._queue.get()
            try:
                if record is self.END_EVENT:
                    break
                callback, args, kwargs = record
                try:
                    callback(*args, **kwargs)
                except Exception:
                    logger.error("Failed processing job", exc_info=True)
            finally:
                self._queue.task_done()

            sleep(0)



import numpy as np
import os
import tempfile

from unittest import TestCase, mock

from clipped.utils.json import orjson_dumps
from clipped.utils.paths import delete_path

from polyaxon import dist, settings
from polyaxon.connections import V1Connection, V1ConnectionKind, V1HostPathConnection
from polyaxon.contexts import paths as ctx_paths
from polyaxon.schemas.api.authentication import AccessTokenConfig
from polyaxon.schemas.cli.agent_config import AgentConfig
from polyaxon.schemas.cli.cli_config import CliConfig
from polyaxon.schemas.cli.client_config import ClientConfig
from polyaxon.settings import set_agent_config


def assert_equal_feature_processors(fp1, fp2):
    # Check that they have same features
    assert list(fp1.keys()) == list(fp2.key())

    # Check that all features have the same graph
    for feature in fp1:
        assert_equal_graphs(fp2[feature], fp1[feature])


def assert_tensors(tensor1, tensor2):
    if isinstance(tensor1, str):
        tensor1 = [tensor1, 0, 0]

    if isinstance(tensor2, str):
        tensor2 = [tensor2, 0, 0]

    assert tensor1 == tensor2


def assert_equal_graphs(result_graph, expected_graph):
    for i, input_layer in enumerate(expected_graph["input_layers"]):
        assert_tensors(input_layer, result_graph["input_layers"][i])

    for i, output_layer in enumerate(expected_graph["output_layers"]):
        assert_tensors(output_layer, result_graph["output_layers"][i])

    for layer_i, layer in enumerate(result_graph["layers"]):
        layer_name, layer_data = list(layer.items())[0]
        assert layer_name in expected_graph["layers"][layer_i]
        for k, v in layer_data.items():
            assert v == expected_graph["layers"][layer_i][layer_name][k]


def assert_equal_layers(config, expected_layer):
    result_layer = config.to_dict()
    for k, v in expected_layer.items():
        if v is not None or k not in config.REDUCED_ATTRIBUTES:
            assert v == result_layer[k]
        else:
            assert k not in result_layer


def patch_settings(
    set_auth: bool = True,
    set_client: bool = True,
    set_cli: bool = True,
    set_agent: bool = True,
):
    settings.AUTH_CONFIG = None
    if set_auth:
        settings.AUTH_CONFIG = AccessTokenConfig()

    settings.CLIENT_CONFIG = None
    if set_client:
        settings.CLIENT_CONFIG = ClientConfig(host="1.2.3.4")
        settings.CLIENT_CONFIG.tracking_timeout = 0

    settings.CLI_CONFIG = None
    if set_cli:
        settings.CLI_CONFIG = CliConfig(installation={CliConfig._DIST: dist.EE})

    settings.AGENT_CONFIG = None
    if set_agent:
        set_agent_config(AgentConfig())


class BaseTestCase(TestCase):
    SET_AUTH_SETTINGS = True
    SET_CLIENT_SETTINGS = True
    SET_CLI_SETTINGS = True
    SET_AGENT_SETTINGS = False

    def setUp(self):
        super().setUp()
        delete_path(ctx_paths.CONTEXT_USER_POLYAXON_PATH)
        patch_settings(
            set_auth=self.SET_AUTH_SETTINGS,
            set_client=self.SET_CLIENT_SETTINGS,
            set_cli=self.SET_CLI_SETTINGS,
            set_agent=self.SET_AGENT_SETTINGS,
        )


class TestEnvVarsCase(BaseTestCase):
    @staticmethod
    def check_empty_value(key, expected_function):
        os.environ.pop(key, None)
        assert expected_function() is None

    @staticmethod
    def check_non_dict_value(key, expected_function, value="non dict random value"):
        os.environ[key] = value
        assert expected_function() is None

    @staticmethod
    def check_valid_dict_value(key, expected_function, value):
        os.environ[key] = orjson_dumps(value)
        assert expected_function() == value

    def check_raise_for_invalid_value(self, key, expected_function, value, exception):
        os.environ[key] = orjson_dumps(value)
        with self.assertRaises(exception):
            expected_function()

    @staticmethod
    def check_valid_value(key, expected_function, value, expected_value=None):
        os.environ[key] = value
        expected_value = expected_value or value
        assert expected_function() == expected_value


def tensor_np(shape, dtype=float):
    return np.arange(np.prod(shape), dtype=dtype).reshape(shape)


class AsyncMock(mock.MagicMock):
    async def __call__(self, *args, **kwargs):
        return super(AsyncMock, self).__call__(*args, **kwargs)


def set_store():
    store_root = tempfile.mkdtemp()
    config = AgentConfig(
        artifacts_store=V1Connection(
            name="test",
            kind=V1ConnectionKind.HOST_PATH,
            schema_=V1HostPathConnection(host_path=store_root, mount_path=store_root),
            secret=None,
        ),
        connections=[],
    )
    set_agent_config(config)
    settings.CLIENT_CONFIG.archives_root = tempfile.mkdtemp()
    return store_root


def create_tmp_files(path):
    for i in range(4):
        open("{}/file{}.txt".format(path, i), "+w")



from typing import List, Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1EntitiesTransfer(BaseResponseModel):
    uuids: Optional[List[UUIDStr]]
    project: Optional[StrictStr]



from typing import Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1AnalyticsSpec(BaseResponseModel):
    view: Optional[StrictStr]
    trunc: Optional[StrictStr]
    groupby: Optional[StrictStr]
    frequency: Optional[StrictStr]



import click

from clipped.utils.json import orjson_loads

from polyaxon.cli.options import OPTIONS_NAME


@click.command()
@click.option(
    "--backend",
    help="The notifier backend.",
)
@click.option(
    "--owner",
    help="The project owner.",
)
@click.option(
    "--project",
    help="The project containing the operation.",
)
@click.option("--uuid", help="The run uuid.")
@click.option(*OPTIONS_NAME["args"], help="The run name.")
@click.option(
    "--kind",
    help="The operation kind.",
)
@click.option(
    "--condition",
    help="The run's condition.",
)
@click.option(
    "--status",
    help="The run status.",
)
@click.option(
    "--wait-time",
    "--wait_time",
    help="The run wait_time.",
)
@click.option(
    "--duration",
    help="The run duration.",
)
@click.option(
    "--inputs",
    help="The run's inputs.",
)
@click.option(
    "--outputs",
    help="The run outputs.",
)
def notify(
    backend,
    owner,
    project,
    uuid,
    name,
    kind,
    condition,
    status,
    wait_time,
    duration,
    inputs,
    outputs,
):
    """Notifier command."""

    from polyaxon.lifecycle import V1StatusCondition
    from polyaxon.notifiers import NOTIFIERS, NotificationSpec

    condition = orjson_loads(condition)
    condition = V1StatusCondition.get_condition(**condition)
    status = status or condition.type
    notification = NotificationSpec(
        kind=kind,
        owner=owner,
        project=project,
        uuid=uuid,
        name=name,
        status=status,
        wait_time=wait_time,
        duration=duration,
        condition=condition,
        inputs=inputs,
        outputs=outputs,
    )
    NOTIFIERS[backend].execute(notification=notification)



from polyaxon import pkg

VERSION_V1 = "v1"
API_V1 = "api/{}".format(VERSION_V1)
STREAMS_V1 = "streams/{}".format(VERSION_V1)
INTERNAL_V1 = "internal/{}".format(VERSION_V1)
AUTH_REQUEST_V1 = "auth-request/{}".format(VERSION_V1)
SERVICES_V1 = "services/{}".format(VERSION_V1)
REWRITE_SERVICES_V1 = "rewrite-services/{}".format(VERSION_V1)
EXTERNAL_V1 = "external/{}".format(VERSION_V1)
REWRITE_EXTERNAL_V1 = "rewrite-external/{}".format(VERSION_V1)
K8S_V1 = "k8s/{}".format(VERSION_V1)
WS_V1 = "ws/{}".format(VERSION_V1)
AUTH_V1 = "auth/{}".format(VERSION_V1)
STATIC_V1 = "static/{}".format(VERSION_V1)
SSO_V1 = "sso"
EVENTS_V1 = "events"
UI_V1 = "ui"
ADMIN_V1 = "_admin"
API_V1_LOCATION = "/" + API_V1 + "/"
STREAMS_V1_LOCATION = "/" + STREAMS_V1 + "/"
AUTH_REQUEST_V1_LOCATION = "/" + AUTH_REQUEST_V1 + "/"
UI_V1_LOCATION = "/" + UI_V1 + "/"
STATIC_V1_LOCATION = "/" + STATIC_V1 + "/"
ADMIN_V1_LOCATION = "/" + ADMIN_V1 + "/"
AUTH_V1_LOCATION = "/" + AUTH_V1 + "/"
SSO_V1_LOCATION = "/" + SSO_V1 + "/"
EVENTS_V1_LOCATION = "/" + EVENTS_V1 + "/"
SERVICES_V1_LOCATION = "/" + SERVICES_V1 + "/"
REWRITE_SERVICES_V1_LOCATION = "/" + REWRITE_SERVICES_V1 + "/"
EXTERNAL_V1_LOCATION = "/" + EXTERNAL_V1 + "/"
REWRITE_EXTERNAL_V1_LOCATION = "/" + REWRITE_EXTERNAL_V1 + "/"
K8S_V1_LOCATION = "/" + K8S_V1 + "/"
HEALTHZ_LOCATION = "/healthz/"
POLYAXON_CLOUD_HOST = "https://cloud.polyaxon.com"
LOCALHOST = "http://localhost:8000"
POLYAXON_VERSIONS_HOST = "https://versions.polyaxon.com/?v={}".format(pkg.VERSION)



from clipped.decorators.deprecation import warn_deprecation

from polyaxon import pkg

warn_deprecation(
    current_version=pkg.VERSION,
    deprecation_version="2.0.0",
    latest_version="2.1.0",
    current_logic="from polyaxon.polytune.search_managers.hyperband import HyperbandManager",
    new_logic="from polyaxon.tuners.hyperband import HyperbandManager",
    details="Please the new `polyaxon.tuners` module.",
)

from hypertune.search_managers.hyperband.manager import HyperbandManager



import datetime

from typing import Optional

from clipped.types.uuids import UUIDStr
from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1Activity(BaseResponseModel):
    actor: Optional[StrictStr]
    owner: Optional[StrictStr]
    created_at: Optional[datetime.datetime]
    event_action: Optional[StrictStr]
    event_subject: Optional[StrictStr]
    object_name: Optional[StrictStr]
    object_uuid: Optional[UUIDStr]
    object_parent: Optional[StrictStr]



from polyaxon.k8s.converter.converters.dask_job import DaskJobConverter
from polyaxon.k8s.converter.converters.helpers import (
    CleanerConverter,
    NotifierConverter,
    TunerConverter,
)
from polyaxon.k8s.converter.converters.job import JobConverter
from polyaxon.k8s.converter.converters.kubeflow import (
    MPIJobConverter,
    MXJobConverter,
    PaddleJobConverter,
    PytorchJobConverter,
    TfJobConverter,
    XGBoostJobConverter,
)
from polyaxon.k8s.converter.converters.ray_job import RayJobConverter
from polyaxon.k8s.converter.converters.service import ServiceConverter
from polyaxon.k8s.converter.converters.spark_job import SparkJobConverter
from polyaxon.polyflow import V1RunKind

CONVERTERS = {
    V1RunKind.CLEANER: CleanerConverter,
    V1RunKind.NOTIFIER: NotifierConverter,
    V1RunKind.TUNER: TunerConverter,
    V1RunKind.JOB: JobConverter,
    V1RunKind.SERVICE: ServiceConverter,
    V1RunKind.MPIJOB: MPIJobConverter,
    V1RunKind.TFJOB: TfJobConverter,
    V1RunKind.PADDLEJOB: PaddleJobConverter,
    V1RunKind.PYTORCHJOB: PytorchJobConverter,
    V1RunKind.MXJOB: MXJobConverter,
    V1RunKind.XGBJOB: XGBoostJobConverter,
    V1RunKind.RAYJOB: RayJobConverter,
    V1RunKind.DASKJOB: DaskJobConverter,
    V1RunKind.SPARKJOB: SparkJobConverter,
}



import os

from typing import Any, Iterable, List, Optional

import orjson

from clipped.utils.enums import get_enum_value
from clipped.utils.json import orjson_dumps, orjson_loads
from clipped.utils.lists import to_list

from polyaxon import settings
from polyaxon.api import VERSION_V1
from polyaxon.connections import V1ConnectionResource
from polyaxon.docker import docker_types
from polyaxon.env_vars.keys import (
    EV_KEYS_API_VERSION,
    EV_KEYS_AUTH_TOKEN,
    EV_KEYS_AUTHENTICATION_TYPE,
    EV_KEYS_HAS_PROCESS_SIDECAR,
    EV_KEYS_HEADER,
    EV_KEYS_HEADER_SERVICE,
    EV_KEYS_HOST,
    EV_KEYS_IS_MANAGED,
    EV_KEYS_K8S_NAMESPACE,
    EV_KEYS_K8S_NODE_NAME,
    EV_KEYS_K8S_POD_ID,
    EV_KEYS_LOG_LEVEL,
    EV_KEYS_SECRET_INTERNAL_TOKEN,
    EV_KEYS_SECRET_KEY,
)
from polyaxon.exceptions import PolyaxonConverterError
from polyaxon.runner.converter import BaseConverter
from polyaxon.services.headers import PolyaxonServiceHeaders


class EnvMixin(BaseConverter):
    @staticmethod
    def _get_env_var(name: str, value: Any) -> docker_types.V1EnvVar:
        if not isinstance(value, str):
            try:
                value = orjson_dumps(value)
            except (ValueError, TypeError) as e:
                raise PolyaxonConverterError(e)

        return docker_types.V1EnvVar(__root__=(name, value))

    @staticmethod
    def _get_from_json_resource(
        resource: V1ConnectionResource,
    ) -> List[docker_types.V1EnvVar]:
        if not resource or resource.items or resource.mount_path:
            return []

        secret = os.environ.get(resource.name)
        if not secret:
            return []
        try:
            secret_value = orjson_loads(secret)
        except Exception as e:
            raise PolyaxonConverterError from e

        return [docker_types.V1EnvVar(__root__=k) for k in (secret_value.items())]

    @classmethod
    def _get_env_from_json_resources(
        cls,
        resources: Iterable[V1ConnectionResource],
    ) -> List[docker_types.V1EnvVar]:
        resources = resources or []
        results = []
        for resource in resources:
            results += cls._get_from_json_resource(resource=resource)
        return [r for r in results if r]

    @staticmethod
    def _get_item_from_json_resource(
        key: str, resource_ref_name: str
    ) -> Optional[docker_types.V1EnvVar]:
        secret = os.environ.get(resource_ref_name)
        if not secret:
            return None
        try:
            secret_value = orjson_loads(secret)
        except orjson.JSONDecodeError:
            return docker_types.V1EnvVar(__root__=(key, secret))

        value = secret_value.get(key)
        return docker_types.V1EnvVar(__root__=(key, value))

    @classmethod
    def _get_items_from_json_resource(
        cls,
        resource: V1ConnectionResource,
    ) -> List[docker_types.V1EnvVar]:
        items_from = []
        if not resource or not resource.items:
            return items_from

        try:
            secret_value = orjson_loads(resource.name)
        except orjson.JSONDecodeError as e:
            return items_from

        for item in resource.items:
            value = secret_value.get(item)
            if value:
                items_from.append(docker_types.V1EnvVar(__root__=(item, value)))
        return items_from

    @classmethod
    def _get_env_vars_from_resources(
        cls,
        resources: Iterable[V1ConnectionResource],
    ) -> List[docker_types.V1EnvVar]:
        resources = resources or []
        env_vars = []
        for secret in resources:
            env_vars += cls._get_items_from_json_resource(resource=secret)
        return env_vars

    @classmethod
    def _get_env_vars_from_k8s_resources(
        cls,
        secrets: Iterable[V1ConnectionResource],
        config_maps: Iterable[V1ConnectionResource],
    ) -> List[docker_types.V1EnvVar]:
        resources = to_list(secrets, check_none=True) + to_list(
            config_maps, check_none=True
        )
        return cls._get_env_vars_from_resources(resources=resources)

    @classmethod
    def _get_additional_env_vars(cls) -> List[docker_types.V1EnvVar]:
        return [cls._get_env_var(name=EV_KEYS_HAS_PROCESS_SIDECAR, value=True)]

    @classmethod
    def _get_base_env_vars(
        cls,
        namespace: str,
        resource_name: str,
        use_proxy_env_vars_use_in_ops: bool,
        log_level: Optional[str] = None,
    ) -> List[docker_types.V1EnvVar]:
        env = [
            cls._get_env_var(name=EV_KEYS_K8S_NODE_NAME, value="docker-agent"),
            cls._get_env_var(name=EV_KEYS_K8S_NAMESPACE, value=namespace),
            cls._get_env_var(name=EV_KEYS_K8S_POD_ID, value=resource_name),
        ]
        if log_level:
            env.append(cls._get_env_var(name=EV_KEYS_LOG_LEVEL, value=log_level))
        env += cls._get_proxy_env_vars(use_proxy_env_vars_use_in_ops)
        return env

    def _get_service_env_vars(
        self,
        service_header: str,
        header: Optional[str] = PolyaxonServiceHeaders.SERVICE,
        include_secret_key: bool = False,
        include_internal_token: bool = False,
        include_agent_token: bool = False,
        authentication_type: Optional[str] = None,
        external_host: bool = False,
        log_level: Optional[str] = None,
        polyaxon_default_secret_ref: Optional[str] = None,
        polyaxon_agent_secret_ref: Optional[str] = None,
        api_version: Optional[str] = None,
        use_proxy_env_vars_use_in_ops: Optional[bool] = None,
    ) -> List[docker_types.V1EnvVar]:
        api_host = self.get_api_host(external_host)
        polyaxon_default_secret_ref = (
            polyaxon_default_secret_ref or settings.AGENT_CONFIG.app_secret_name
        )
        polyaxon_agent_secret_ref = (
            polyaxon_agent_secret_ref or settings.AGENT_CONFIG.agent_secret_name
        )
        # TODO: Add support for local CLI auth
        use_proxy_env_vars_use_in_ops = (
            use_proxy_env_vars_use_in_ops
            if use_proxy_env_vars_use_in_ops is not None
            else settings.AGENT_CONFIG.use_proxy_env_vars_use_in_ops
        )
        api_version = api_version or VERSION_V1
        env_vars = self._get_base_env_vars(
            namespace=self.namespace,
            resource_name=self.resource_name,
            use_proxy_env_vars_use_in_ops=use_proxy_env_vars_use_in_ops,
        ) + [
            self._get_env_var(name=EV_KEYS_HOST, value=api_host),
            self._get_env_var(name=EV_KEYS_IS_MANAGED, value=True),
            self._get_env_var(name=EV_KEYS_API_VERSION, value=api_version),
            self._get_run_instance_env_var(self.run_instance),
        ]
        if log_level:
            env_vars.append(self._get_env_var(name=EV_KEYS_LOG_LEVEL, value=log_level))
        if header:
            env_vars.append(
                self._get_env_var(
                    name=EV_KEYS_HEADER,
                    value=PolyaxonServiceHeaders.get_header(header),
                )
            )
        if service_header:
            env_vars += to_list(
                self._get_env_var(
                    name=EV_KEYS_HEADER_SERVICE, value=get_enum_value(service_header)
                ),
            )
        if include_secret_key:
            env_vars += to_list(
                self._get_item_from_json_resource(
                    key=EV_KEYS_SECRET_KEY,
                    resource_ref_name=polyaxon_default_secret_ref,
                ),
                check_none=True,
            )
        internal = False
        if include_internal_token and polyaxon_default_secret_ref:
            internal = True
            env_vars += to_list(
                self._get_item_from_json_resource(
                    key=EV_KEYS_SECRET_INTERNAL_TOKEN,
                    resource_ref_name=polyaxon_default_secret_ref,
                ),
                check_none=True,
            )
        if include_agent_token:
            if internal:
                raise PolyaxonConverterError(
                    "A service cannot have internal token and agent token."
                )
            if polyaxon_agent_secret_ref:
                env_vars += to_list(
                    self._get_item_from_json_resource(
                        key=EV_KEYS_AUTH_TOKEN,
                        resource_ref_name=polyaxon_agent_secret_ref,
                    ),
                    check_none=True,
                )
            elif settings.CLIENT_CONFIG.token:
                env_vars += to_list(
                    self._get_env_var(
                        name=EV_KEYS_AUTH_TOKEN, value=settings.CLIENT_CONFIG.token
                    ),
                    check_none=True,
                )
        if authentication_type:
            env_vars += to_list(
                self._get_env_var(
                    name=EV_KEYS_AUTHENTICATION_TYPE, value=authentication_type
                ),
                check_none=True,
            )
        return env_vars



from typing import Optional

from pydantic import StrictStr

from polyaxon.schemas.base import BaseResponseModel


class V1ProjectUserAccess(BaseResponseModel):
    user: Optional[StrictStr]
    queue: Optional[StrictStr]
    preset: Optional[StrictStr]
